# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2015, 2016, 2017 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
minio.api
~~~~~~~~~~~~

This module implements the API.

:copyright: (c) 2015, 2016, 2017 by MinIO, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

# Standard python packages
from __future__ import absolute_import
import platform
from threading import Thread

from datetime import datetime, timedelta

import io
import json
import os
import itertools


try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError

# Dependencies
import urllib3
import certifi
import dateutil.parser

# Internal imports
from . import __title__, __version__
from .compat import (urlsplit, queryencode,
                     range, basestring)
from .error import (ResponseError, NoSuchBucket, AccessDenied,
                    InvalidArgumentError, InvalidSizeError, InvalidXMLError)
from .definitions import Object, UploadPart
from .parsers import (parse_list_buckets,
                      parse_list_objects,
                      parse_list_objects_v2,
                      parse_list_parts,
                      parse_copy_object,
                      parse_list_multipart_uploads,
                      parse_new_multipart_upload,
                      parse_location_constraint,
                      parse_multipart_upload_result,
                      parse_get_bucket_notification,
                      parse_multi_object_delete_response)
from .helpers import (get_target_url, is_non_empty_string,
                      is_valid_endpoint,
                      get_sha256_hexdigest, get_md5_base64digest,
                      optimal_part_info,
                      is_valid_bucket_name, read_full,
                      get_s3_region_from_endpoint, is_valid_sse_object, is_valid_sse_c_object,
                      is_valid_source_sse_object,
                      is_valid_bucket_notification_config, is_valid_policy_type,
                      mkdir_p, dump_http, amzprefix_user_metadata,
                      is_supported_header,is_amz_header)
from .helpers import (MAX_PART_SIZE,
                      MAX_POOL_SIZE,
                      MIN_PART_SIZE,
                      DEFAULT_PART_SIZE,
                      MAX_MULTIPART_COUNT)
from .signer import (sign_v4, presign_v4,
                     generate_credential_string,
                     post_presign_signature)
from .signer import (_UNSIGNED_PAYLOAD, _SIGN_V4_ALGORITHM)
from .xml_marshal import (xml_marshal_bucket_constraint,
                          xml_marshal_complete_multipart_upload,
                          xml_marshal_bucket_notifications,
                          xml_marshal_delete_objects,
                          xml_marshal_select)
from .fold_case_dict import FoldCaseDict
from .thread_pool import ThreadPool
from .select import SelectObjectReader

# Comment format.
_COMMENTS = '({0}; {1})'
# App info format.
_APP_INFO = '{0}/{1}'

# MinIO (OS; ARCH) LIB/VER APP/VER .
_DEFAULT_USER_AGENT = 'MinIO {0} {1}'.format(
    _COMMENTS.format(platform.system(),
                     platform.machine()),
    _APP_INFO.format(__title__,
                     __version__))


# Duration of 7 days in seconds
_MAX_EXPIRY_TIME = 604800 # 7 days in seconds

# Number of parallel workers which upload parts
_PARALLEL_UPLOADERS = 3


class Minio(object):
    """
    Constructs a :class:`Minio <Minio>`.

    Examples:
        client = Minio('play.min.io')
        client = Minio('s3.amazonaws.com', 'ACCESS_KEY', 'SECRET_KEY')

        # To override auto bucket location discovery.
        client = Minio('play.min.io', 'ACCESS_KEY', 'SECRET_KEY',
                       region='us-east-1')

    :param endpoint: Hostname of the cloud storage server.
    :param access_key: Access key to sign self._http.request with.
    :param secret_key: Secret key to sign self._http.request with.
    :param session_token: Session token to sign self._http.request with.
    :param secure: Set this value if wish to make secure requests.
         Default is True.
    :param region: Set this value to override automatic bucket
         location discovery.
    :param timeout: Set this value to control how long requests
         are allowed to run before being aborted.
    :return: :class:`Minio <Minio>` object
    """
    def __init__(self, endpoint, access_key=None,
                 secret_key=None,
                 session_token=None,
                 secure=True,
                 region=None,
                 http_client=None):

        # Validate endpoint.
        is_valid_endpoint(endpoint)

        # Validate http client has correct base class.
        if http_client and not isinstance(http_client, urllib3.poolmanager.PoolManager):
            raise InvalidArgumentError('HTTP client should be of instance `urllib3.poolmanager.PoolManager`')


        # Default is a secured connection.
        endpoint_url = 'https://' + endpoint
        if not secure:
            endpoint_url = 'http://' + endpoint

        # Parse url endpoints.
        url_components = urlsplit(endpoint_url)

        # Extract region if possible from endpoint.
        if not region:
            region = get_s3_region_from_endpoint(endpoint)

        self._region = region
        self._region_map = dict()
        self._endpoint_url = url_components.geturl()
        self._is_ssl = secure
        self._access_key = access_key
        self._secret_key = secret_key
        self._session_token = session_token
        self._user_agent = _DEFAULT_USER_AGENT
        self._trace_output_stream = None

        # Load CA certificates from SSL_CERT_FILE file if set
        ca_certs = os.environ.get('SSL_CERT_FILE')
        if not ca_certs:
            ca_certs = certifi.where()

        if not http_client:
            self._http = urllib3.PoolManager(
                timeout=urllib3.Timeout.DEFAULT_TIMEOUT,
                maxsize=MAX_POOL_SIZE,
                        cert_reqs='CERT_REQUIRED',
                        ca_certs=ca_certs,
                        retries=urllib3.Retry(
                            total=5,
                            backoff_factor=0.2,
                            status_forcelist=[500, 502, 503, 504]
                        )
            )
        else:
            self._http = http_client

    # Set application information.
    def set_app_info(self, app_name, app_version):
        """
        Sets your application name and version to
        default user agent in the following format.

              MinIO (OS; ARCH) LIB/VER APP/VER

        Example:
            client.set_app_info('my_app', '1.0.2')

        :param app_name: application name.
        :param app_version: application version.
        """
        if not (app_name and app_version):
            raise ValueError('app_name and app_version cannot be empty.')

        app_info = _APP_INFO.format(app_name,
                                    app_version)
        self._user_agent = ' '.join([_DEFAULT_USER_AGENT, app_info])

    # enable HTTP trace.
    def trace_on(self, stream):
        """
        Enable http trace.

        :param output_stream: Stream where trace is written to.
        """
        if not stream:
            raise ValueError('Input stream for trace output is invalid.')
        # Save new output stream.
        self._trace_output_stream = stream

    # disable HTTP trace.
    def trace_off(self):
        """
        Disable HTTP trace.
        """
        self._trace_output_stream = None

    # Select Object Content
    def select_object_content(self, bucket_name, object_name, opts):
        """
        Executes SQL requests on objects having data in CSV, JSON
        or Parquet formats.

        Examples:
            data = client.select_object_content('foo', 'test.csv', options)

        :param bucket_name: Bucket to read object from
        :param object_name: Name of object to read
        :param options: Options for select object
        """
        is_valid_bucket_name(bucket_name)
        is_non_empty_string(object_name)

        content = xml_marshal_select(opts)
        url_values = dict()
        url_values["select"] = ""
        url_values["select-type"] = "2"

        headers = {
            'Content-Length': str(len(content)),
            'Content-Md5': get_md5_base64digest(content)
        }
        content_sha256_hex = get_sha256_hexdigest(content)
        response = self._url_open(
            'POST',
            bucket_name=bucket_name,
            object_name=object_name,
            query=url_values,
            headers=headers,
            body=content,
            content_sha256=content_sha256_hex,
            preload_content=False)

        return SelectObjectReader(response)

    # Bucket level
    def make_bucket(self, bucket_name, location='us-east-1'):
        """
        Make a new bucket on the server.

        Optionally include Location.
           ['us-east-1', 'us-east-2', 'us-west-1', 'us-west-2', 'eu-west-1',
            'eu-west-2', 'ca-central-1', 'eu-central-1', 'sa-east-1',
            'cn-north-1', 'ap-southeast-1', 'ap-southeast-2',
            'ap-northeast-1', 'ap-northeast-2']

        Examples:
            minio.make_bucket('foo')
            minio.make_bucket('foo', 'us-west-1')

        :param bucket_name: Bucket to create on server
        :param location: Location to create bucket on
        """
        is_valid_bucket_name(bucket_name)

        # Default region for all requests.
        region = 'us-east-1'
        if self._region:
            region = self._region
            # Validate if caller requested bucket location is same as current region
            if self._region != location:
                raise InvalidArgumentError("Configured region {0}, requested"
                                           " {1}".format(self._region,
                                                         location))

        method = 'PUT'
        # Set user agent once before the request.
        headers = {'User-Agent': self._user_agent}

        content = None
        if location and location != 'us-east-1':
            content = xml_marshal_bucket_constraint(location)
            headers['Content-Length'] = str(len(content))

        content_sha256_hex = get_sha256_hexdigest(content)
        if content:
            headers['Content-Md5'] = get_md5_base64digest(content)

        # In case of Amazon S3.  The make bucket issued on already
        # existing bucket would fail with 'AuthorizationMalformed'
        # error if virtual style is used. So we default to 'path
        # style' as that is the preferred method here. The final
        # location of the 'bucket' is provided through XML
        # LocationConstraint data with the request.
        # Construct target url.
        url = self._endpoint_url + '/' + bucket_name + '/'

        # Get signature headers if any.
        headers = sign_v4(method, url, region,
                          headers, self._access_key,
                          self._secret_key,
                          self._session_token,
                          content_sha256_hex,
                          datetime.utcnow())

        response = self._http.urlopen(method, url,
                                      body=content,
                                      headers=headers)

        if response.status != 200:
            raise ResponseError(response, method, bucket_name).get_exception()

        self._set_bucket_region(bucket_name, region=location)

    def list_buckets(self):
        """
        List all buckets owned by the user.

        Example:
            bucket_list = minio.list_buckets()
            for bucket in bucket_list:
                print(bucket.name, bucket.created_date)

        :return: An iterator of buckets owned by the current user.
        """

        method = 'GET'
        url = get_target_url(self._endpoint_url)
        # Set user agent once before the request.
        headers = {'User-Agent': self._user_agent}

        # default for all requests.
        region = 'us-east-1'
        # region is set then use the region.
        if self._region:
            region = self._region

        # Get signature headers if any.
        headers = sign_v4(method, url, region,
                          headers, self._access_key,
                          self._secret_key,
                          self._session_token,
                          None, datetime.utcnow())

        response = self._http.urlopen(method, url,
                                      body=None,
                                      headers=headers)

        if self._trace_output_stream:
            dump_http(method, url, headers, response,
                      self._trace_output_stream)

        if response.status != 200:
            raise ResponseError(response, method).get_exception()
        try:
            return parse_list_buckets(response.data)
        except InvalidXMLError:
            if self._endpoint_url.endswith("s3.amazonaws.com") and (not self._access_key or not self._secret_key):
                raise AccessDenied(response)

    def bucket_exists(self, bucket_name):
        """
        Check if the bucket exists and if the user has access to it.

        :param bucket_name: To test the existence and user access.
        :return: True on success.
        """
        is_valid_bucket_name(bucket_name)

        try:
            self._url_open('HEAD', bucket_name=bucket_name)
        # If the bucket has not been created yet, MinIO will return a "NoSuchBucket" error.
        except NoSuchBucket:
            return False
        except ResponseError:
            raise
        return True

    def remove_bucket(self, bucket_name):
        """
        Remove a bucket.

        :param bucket_name: Bucket to remove
        """
        is_valid_bucket_name(bucket_name)
        self._url_open('DELETE', bucket_name=bucket_name)

        # Make sure to purge bucket_name from region cache.
        self._delete_bucket_region(bucket_name)

    def get_bucket_policy(self, bucket_name):
        """
        Get bucket policy of given bucket name.

        :param bucket_name: Bucket name.
        """
        is_valid_bucket_name(bucket_name)

        response = self._url_open("GET",
                                  bucket_name=bucket_name,
                                  query={"policy": ""})
        return response.data

    def delete_bucket_policy(self, bucket_name):
        self._url_open("DELETE",
                        bucket_name=bucket_name,
                        query={"policy": ""})

    def set_bucket_policy(self, bucket_name, policy):
        """
        Set bucket policy of given bucket name.

        :param bucket_name: Bucket name.
        :param policy: Access policy/ies in string format.
        """
        is_valid_policy_type(policy)

        is_valid_bucket_name(bucket_name)

        headers = {
            'Content-Length': str(len(policy)),
            'Content-Md5': get_md5_base64digest(policy)
        }
        content_sha256_hex = get_sha256_hexdigest(policy)
        self._url_open("PUT",
                        bucket_name=bucket_name,
                        query={"policy": ""},
                        headers=headers,
                        body=policy,
                        content_sha256=content_sha256_hex)

    def get_bucket_notification(self, bucket_name):
        """
        Get notifications configured for the given bucket.

        :param bucket_name: Bucket name.
        """
        is_valid_bucket_name(bucket_name)

        response = self._url_open(
            "GET",
            bucket_name=bucket_name,
            query={"notification": ""},
        )
        data = response.data.decode('utf-8')
        return parse_get_bucket_notification(data)

    def set_bucket_notification(self, bucket_name, notifications):
        """
        Set the given notifications on the bucket.

        :param bucket_name: Bucket name.
        :param notifications: Notifications structure
        """
        is_valid_bucket_name(bucket_name)
        is_valid_bucket_notification_config(notifications)

        content = xml_marshal_bucket_notifications(notifications)
        headers = {
            'Content-Length': str(len(content)),
            'Content-Md5': get_md5_base64digest(content)
        }
        content_sha256_hex = get_sha256_hexdigest(content)
        self._url_open(
            'PUT',
            bucket_name=bucket_name,
            query={"notification": ""},
            headers=headers,
            body=content,
            content_sha256=content_sha256_hex
        )

    def remove_all_bucket_notification(self, bucket_name):
        """
        Removes all bucket notification configs configured
        previously, this call disable event notifications
        on a bucket. This operation cannot be undone, to
        set notifications again you should use
        ``set_bucket_notification``

        :param bucket_name: Bucket name.
        """
        is_valid_bucket_name(bucket_name)

        content_bytes = xml_marshal_bucket_notifications({})
        headers = {
            'Content-Length': str(len(content_bytes)),
            'Content-Md5': get_md5_base64digest(content_bytes)
        }
        content_sha256_hex = get_sha256_hexdigest(content_bytes)
        self._url_open(
            'PUT',
            bucket_name=bucket_name,
            query={"notification": ""},
            headers=headers,
            body=content_bytes,
            content_sha256=content_sha256_hex
        )

    def listen_bucket_notification(self, bucket_name, prefix='', suffix='',
                                   events=['s3:ObjectCreated:*',
                                           's3:ObjectRemoved:*',
                                           's3:ObjectAccessed:*']):
        """
        Yeilds new event notifications on a bucket, caller should iterate
        to read new notifications.

        NOTE: Notification is retried in case of `JSONDecodeError` otherwise
        the function raises an exception.

        :param bucket_name: Bucket name to listen event notifications from.
        :param prefix: Object key prefix to filter notifications for.
        :param suffix: Object key suffix to filter notifications for.
        :param events: Enables notifications for specific event types.
             of events.
        """
        is_valid_bucket_name(bucket_name)

        # If someone explicitly set prefix to None convert it to empty string.
        if prefix is None:
            prefix = ''

        # If someone explicitly set suffix to None convert it to empty string.
        if suffix is None:
            suffix = ''

        url_components = urlsplit(self._endpoint_url)
        if url_components.hostname == 's3.amazonaws.com':
            raise InvalidArgumentError(
                'Listening for event notifications on a bucket is a MinIO '
                'specific extension to bucket notification API. It is not '
                'supported by Amazon S3')

        query = {
            'prefix': prefix,
            'suffix': suffix,
            'events': events,
        }
        while True:
            response = self._url_open('GET', bucket_name=bucket_name,
                                      query=query, preload_content=False)
            try:
                for line in response.stream():
                    if line.strip():
                        if hasattr(line, 'decode'):
                            line = line.decode('utf-8')
                        event = json.loads(line)
                        if event['Records'] is not None:
                            yield event
            except JSONDecodeError:
                response.close()
                continue

    def fput_object(self, bucket_name, object_name, file_path,
                    content_type='application/octet-stream',
                    metadata=None, sse=None, progress=None,
                    part_size=DEFAULT_PART_SIZE):
        """
        Add a new object to the cloud storage server.

        Examples:
            minio.fput_object('foo', 'bar', 'filepath', 'text/plain')

        :param bucket_name: Bucket to read object from.
        :param object_name: Name of the object to read.
        :param file_path: Local file path to be uploaded.
        :param content_type: Content type of the object.
        :param metadata: Any additional metadata to be uploaded along
            with your PUT request.
        :param progress: A progress object
        :param part_size: Multipart part size
        :return: etag
        """

        # Open file in 'read' mode.
        with open(file_path, 'rb') as file_data:
            file_size = os.stat(file_path).st_size
            return self.put_object(bucket_name, object_name, file_data,
                                   file_size, content_type, metadata, sse,
                                   progress, part_size)

    def fget_object(self, bucket_name, object_name, file_path, request_headers=None, sse=None):
        """
        Retrieves an object from a bucket and writes at file_path.

        Examples:
            minio.fget_object('foo', 'bar', 'localfile')

        :param bucket_name: Bucket to read object from.
        :param object_name: Name of the object to read.
        :param file_path: Local file path to save the object.
        :param request_headers: Any additional headers to be added with GET request.
        """
        is_valid_bucket_name(bucket_name)
        is_non_empty_string(object_name)

        stat = self.stat_object(bucket_name, object_name, sse)

        if os.path.isdir(file_path):
            raise OSError("file is a directory.")

        # Create top level directory if needed.
        top_level_dir = os.path.dirname(file_path)
        if top_level_dir:
            mkdir_p(top_level_dir)

        # Write to a temporary file "file_path.part.minio" before saving.
        file_part_path = file_path + stat.etag + '.part.minio'

        # Open file in 'overwrite' mode.
        with open(file_part_path, 'wb') as file_part_data:
            # Save current file_part statinfo.
            file_statinfo = os.stat(file_part_path)

            # Get partial object.
            response = self._get_partial_object(bucket_name, object_name,
                                                offset=file_statinfo.st_size,
                                                length=0,
                                                request_headers=request_headers,
                                                sse=sse)

            # Save content_size to verify if we wrote more data.
            content_size = int(response.headers['content-length'])

            # Save total_written.
            total_written = 0
            for data in response.stream(amt=1024 * 1024):
                file_part_data.write(data)
                total_written += len(data)

            # Release the connection from the response at this point.
            response.release_conn()

            # Verify if we wrote data properly.
            if total_written < content_size:
                msg = 'Data written {0} bytes is smaller than the ' \
                      'specified size {1} bytes'.format(total_written,
                                                        content_size)
                raise InvalidSizeError(msg)

            if total_written > content_size:
                msg = 'Data written {0} bytes is in excess than the ' \
                      'specified size {1} bytes'.format(total_written,
                                                        content_size)
                raise InvalidSizeError(msg)

        #Delete existing file to be compatible with Windows
        if os.path.exists(file_path):
               os.remove(file_path)
        #Rename with destination file path
        os.rename(file_part_path, file_path)

        # Return the stat
        return stat

    def get_object(self, bucket_name, object_name, request_headers=None, sse=None):
        """
        Retrieves an object from a bucket.

        This function returns an object that contains an open network
        connection to enable incremental consumption of the
        response. To re-use the connection (if desired) on subsequent
        requests, the user needs to call `release_conn()` on the
        returned object after processing.

        Examples:
            my_object = minio.get_partial_object('foo', 'bar')

        :param bucket_name: Bucket to read object from
        :param object_name: Name of object to read
        :param request_headers: Any additional headers to be added with GET request.
        :return: :class:`urllib3.response.HTTPResponse` object.

        """
        is_valid_bucket_name(bucket_name)
        is_non_empty_string(object_name)

        return self._get_partial_object(bucket_name,
                                        object_name,
                                        request_headers=request_headers,
                                        sse=sse)

    def get_partial_object(self, bucket_name, object_name, offset=0, length=0, request_headers=None, sse=None):
        """
        Retrieves an object from a bucket.

        Optionally takes an offset and length of data to retrieve.

        This function returns an object that contains an open network
        connection to enable incremental consumption of the
        response. To re-use the connection (if desired) on subsequent
        requests, the user needs to call `release_conn()` on the
        returned object after processing.

        Examples:
            partial_object = minio.get_partial_object('foo', 'bar', 2, 4)

        :param bucket_name: Bucket to retrieve object from
        :param object_name: Name of object to retrieve
        :param offset: Optional offset to retrieve bytes from.
           Must be >= 0.
        :param length: Optional number of bytes to retrieve.
           Must be an integer.
        :param request_headers: Any additional headers to be added with GET request.
        :return: :class:`urllib3.response.HTTPResponse` object.

        """
        is_valid_bucket_name(bucket_name)
        is_non_empty_string(object_name)

        return self._get_partial_object(bucket_name,
                                        object_name,
                                        offset, length,
                                        request_headers=request_headers,
                                        sse=sse)

    def copy_object(self, bucket_name, object_name, object_source,
                    conditions=None, source_sse=None, sse=None, metadata=None):
        """
        Copy a source object on object storage server to a new object.

        NOTE: Maximum object size supported by this API is 5GB.

        Examples:

        :param bucket_name: Bucket of new object.
        :param object_name: Name of new object.
        :param object_source: Source object to be copied.
        :param conditions: :class:`CopyConditions` object. Collection of
        supported CopyObject conditions.
        :param metadata: Any user-defined metadata to be copied along with
        destination object.
        """
        is_valid_bucket_name(bucket_name)
        is_non_empty_string(object_name)
        is_non_empty_string(object_source)

        headers = {}

        # Preserving the user-defined metadata in headers
        if metadata is not None:
            headers = amzprefix_user_metadata(metadata)
            headers["x-amz-metadata-directive"] = "REPLACE"

        if conditions:
            for k, v in conditions.items():
                headers[k] = v

        # Source argument to copy_object can only be of type copy_SSE_C
        if source_sse:
            is_valid_source_sse_object(source_sse)
            headers.update(source_sse.marshal())

        #Destination argument to copy_object cannot be of type copy_SSE_C
        if sse:
            is_valid_sse_object(sse)
            headers.update(sse.marshal())

        headers['X-Amz-Copy-Source'] = queryencode(object_source)
        response = self._url_open('PUT',
                                  bucket_name=bucket_name,
                                  object_name=object_name,
                                  headers=headers)

        return parse_copy_object(bucket_name, object_name, response.data)

    def put_object(self, bucket_name, object_name, data, length,
                   content_type='application/octet-stream',
                   metadata=None, sse=None, progress=None,
                   part_size=DEFAULT_PART_SIZE):
        """
        Add a new object to the cloud storage server.

        NOTE: Maximum object size supported by this API is 5TiB.

        Examples:
         file_stat = os.stat('hello.txt')
         with open('hello.txt', 'rb') as data:
             minio.put_object('foo', 'bar', data, file_stat.st_size, 'text/plain')

        - For length lesser than 5MB put_object automatically
          does single Put operation.
        - For length larger than 5MB put_object automatically
          does resumable multipart operation.

        :param bucket_name: Bucket of new object.
        :param object_name: Name of new object.
        :param data: Contents to upload.
        :param length: Total length of object.
        :param content_type: mime type of object as a string.
        :param metadata: Any additional metadata to be uploaded along
            with your PUT request.
        :param progress: A progress object
        :param part_size: Multipart part size
        :return: etag
        """

        is_valid_sse_object(sse)
        is_valid_bucket_name(bucket_name)
        is_non_empty_string(object_name)

        if progress:
            if not isinstance(progress, Thread):
                raise TypeError('Progress object should inherit the thread.')
            # Set progress bar length and object name before upload
            progress.set_meta(total_length=length, object_name=object_name)

        if not callable(getattr(data, 'read')):
            raise ValueError(
                'Invalid input data does not implement a callable read() method')

        if length > (part_size * MAX_MULTIPART_COUNT):
            raise InvalidArgumentError('Part size * max_parts(10000) is '
                                       ' lesser than input length.')

        if part_size < MIN_PART_SIZE:
            raise InvalidArgumentError('Input part size is smaller '
                                       ' than allowed minimum of 5MiB.')

        if part_size > MAX_PART_SIZE:
            raise InvalidArgumentError('Input part size is bigger '
                                       ' than allowed maximum of 5GiB.')

        if not metadata:
            metadata = {}

        metadata = amzprefix_user_metadata(metadata)

        metadata['Content-Type'] = 'application/octet-stream' if \
            not content_type else content_type

        if length > part_size:
            return self._stream_put_object(bucket_name, object_name,
                                           data, length, metadata=metadata,
                                           sse=sse, progress=progress,
                                           part_size=part_size)

        current_data = data.read(length)
        if len(current_data) != length:
            raise InvalidArgumentError(
                'Could not read {} bytes from data to upload'.format(length)
            )

        return self._do_put_object(bucket_name, object_name,
                                   current_data, len(current_data),
                                   metadata=metadata, sse=sse,
                                   progress=progress)

    def list_objects(self, bucket_name, prefix='', recursive=False):
        """
        List objects in the given bucket.

        Examples:
            objects = minio.list_objects('foo')
            for current_object in objects:
                print(current_object)
            # hello
            # hello/
            # hello/
            # world/

            objects = minio.list_objects('foo', prefix='hello/')
            for current_object in objects:
                print(current_object)
            # hello/world/

            objects = minio.list_objects('foo', recursive=True)
            for current_object in objects:
                print(current_object)
            # hello/world/1
            # world/world/2
            # ...

            objects = minio.list_objects('foo', prefix='hello/',
                                         recursive=True)
            for current_object in objects:
                print(current_object)
            # hello/world/1
            # hello/world/2

        :param bucket_name: Bucket to list objects from
        :param prefix: String specifying objects returned must begin with
        :param recursive: If yes, returns all objects for a specified prefix
        :return: An iterator of objects in alphabetical order.
        """
        is_valid_bucket_name(bucket_name)

        # If someone explicitly set prefix to None convert it to empty string.
        if prefix is None:
            prefix = ''

        method = 'GET'

        # Initialize query parameters.
        query = {
            'prefix': prefix
        }

        # Delimited by default.
        if not recursive:
            query['delimiter'] = '/'

        marker = ''
        is_truncated = True
        while is_truncated:
            if marker:
                query['marker'] = marker
            headers = {}
            response = self._url_open(method,
                                      bucket_name=bucket_name,
                                      query=query,
                                      headers=headers)
            objects, is_truncated, marker = parse_list_objects(response.data,
                                                               bucket_name=bucket_name)
            for obj in objects:
                yield obj

    def list_objects_v2(self, bucket_name, prefix='', recursive=False, start_after=''):
        """
        List objects in the given bucket using the List objects V2 API.

        Examples:
            objects = minio.list_objects_v2('foo')
            for current_object in objects:
                print(current_object)
            # hello
            # hello/
            # hello/
            # world/

            objects = minio.list_objects_v2('foo', prefix='hello/')
            for current_object in objects:
                print(current_object)
            # hello/world/

            objects = minio.list_objects_v2('foo', recursive=True)
            for current_object in objects:
                print(current_object)
            # hello/world/1
            # world/world/2
            # ...

            objects = minio.list_objects_v2('foo', prefix='hello/',
                                         recursive=True)
            for current_object in objects:
                print(current_object)
            # hello/world/1
            # hello/world/2


            objects = minio.list_objects_v2('foo', recursive=True,
                                          start_after='hello/world/1')
            for current_object in objects:
                print(current_object)
            # hello/world/2

        :param bucket_name: Bucket to list objects from
        :param prefix: String specifying objects returned must begin with
        :param recursive: If yes, returns all objects for a specified prefix
        :return: An iterator of objects in alphabetical order.
        """
        is_valid_bucket_name(bucket_name)

        # If someone explicitly set prefix to None convert it to empty string.
        if prefix is None:
            prefix = ''

        if start_after is None:
            start_after = ''

        # Initialize query parameters.
        query = {
            'list-type': '2',
            'start-after': start_after,
            'prefix': prefix
        }

        # Delimited by default.
        if not recursive:
            query['delimiter'] = '/'

        continuation_token = None
        is_truncated = True
        while is_truncated:
            if continuation_token is not None:
                query['continuation-token'] = continuation_token
            response = self._url_open(method='GET',
                                      bucket_name=bucket_name,
                                      query=query)
            objects, is_truncated, continuation_token = parse_list_objects_v2(
                response.data, bucket_name=bucket_name
            )

            for obj in objects:
                yield obj

    def stat_object(self, bucket_name, object_name, sse=None):
        """
        Check if an object exists.

        :param bucket_name: Bucket of object.
        :param object_name: Name of object
        :return: Object metadata if object exists
        """

        headers = {}
        if sse:
            is_valid_sse_c_object(sse=sse)
            headers.update(sse.marshal())

        is_valid_bucket_name(bucket_name)
        is_non_empty_string(object_name)

        response = self._url_open('HEAD', bucket_name=bucket_name,
                                  object_name=object_name, headers=headers)

        etag = response.headers.get('etag', '').replace('"', '')
        size = int(response.headers.get('content-length', '0'))
        content_type = response.headers.get('content-type', '')
        last_modified = response.headers.get('last-modified')

        ## Capture only custom metadata.
        custom_metadata = dict()
        for k in response.headers:
            if is_supported_header(k) or is_amz_header(k):
                custom_metadata[k] = response.headers.get(k)

        if last_modified:
            last_modified = dateutil.parser.parse(last_modified).timetuple()
        return Object(bucket_name, object_name, last_modified, etag, size,
                      content_type=content_type, metadata=custom_metadata)

    def remove_object(self, bucket_name, object_name):
        """
        Remove an object from the bucket.

        :param bucket_name: Bucket of object to remove
        :param object_name: Name of object to remove
        :return: None
        """
        is_valid_bucket_name(bucket_name)
        is_non_empty_string(object_name)

        # No reason to store successful response, for errors
        # relevant exceptions are thrown.
        self._url_open('DELETE', bucket_name=bucket_name,
                       object_name=object_name)

    def _process_remove_objects_batch(self, bucket_name, objects_batch):
        """
        Requester and response parser for remove_objects
        """
        # assemble request content for objects_batch
        content = xml_marshal_delete_objects(objects_batch)

        # compute headers
        headers = {
            'Content-Md5': get_md5_base64digest(content),
            'Content-Length': len(content)
        }
        query = {'delete': ''}
        content_sha256_hex = get_sha256_hexdigest(content)

        # send multi-object delete request
        response = self._url_open(
            'POST', bucket_name=bucket_name,
            headers=headers, body=content,
            query=query, content_sha256=content_sha256_hex,
        )

        # parse response to find delete errors
        return parse_multi_object_delete_response(response.data)

    def remove_objects(self, bucket_name, objects_iter):
        """
        Removes multiple objects from a bucket.

        :param bucket_name: Bucket from which to remove objects

        :param objects_iter: A list, tuple or iterator that provides
        objects names to delete.

        :return: An iterator of MultiDeleteError instances for each
        object that had a delete error.

        """
        is_valid_bucket_name(bucket_name)
        if isinstance(objects_iter, basestring):
            raise TypeError(
                'objects_iter cannot be `str` or `bytes` instance. It must be '
                'a list, tuple or iterator of object names'
            )

        # turn list like objects into an iterator.
        objects_iter = itertools.chain(objects_iter)

        obj_batch = []
        exit_loop = False
        while not exit_loop:
            try:
                object_name = next(objects_iter)
                is_non_empty_string(object_name)
            except StopIteration:
                exit_loop = True

            if not exit_loop:
                obj_batch.append(object_name)

            # if we have 1000 items in the batch, or we have to exit
            # the loop, we have to make a request to delete objects.
            if len(obj_batch) == 1000 or (exit_loop and len(obj_batch) > 0):
                # send request and parse response
                errs_result = self._process_remove_objects_batch(
                    bucket_name, obj_batch
                )

                # return the delete errors.
                for err_result in errs_result:
                    yield err_result

                # clear batch for next set of items
                obj_batch = []

    def list_incomplete_uploads(self, bucket_name, prefix='',
                                recursive=False):
        """
        List all in-complete uploads for a given bucket.

        Examples:
            incomplete_uploads = minio.list_incomplete_uploads('foo')
            for current_upload in incomplete_uploads:
                print(current_upload)
            # hello
            # hello/
            # hello/
            # world/

            incomplete_uploads = minio.list_incomplete_uploads('foo',
                                                               prefix='hello/')
            for current_upload in incomplete_uploads:
                print(current_upload)
            # hello/world/

            incomplete_uploads = minio.list_incomplete_uploads('foo',
                                                               recursive=True)
            for current_upload in incomplete_uploads:
                print(current_upload)
            # hello/world/1
            # world/world/2
            # ...

            incomplete_uploads = minio.list_incomplete_uploads('foo',
                                                               prefix='hello/',
                                                               recursive=True)
            for current_upload in incomplete_uploads:
                print(current_upload)
            # hello/world/1
            # hello/world/2

        :param bucket_name: Bucket to list incomplete uploads
        :param prefix: String specifying objects returned must begin with.
        :param recursive: If yes, returns all incomplete uploads for
           a specified prefix.
        :return: An generator of incomplete uploads in alphabetical order.
        """
        is_valid_bucket_name(bucket_name)

        return self._list_incomplete_uploads(bucket_name, prefix, recursive)

    def _list_incomplete_uploads(self, bucket_name, prefix='',
                                 recursive=False, is_aggregate_size=True):
        """
        List incomplete uploads list all previously uploaded incomplete multipart objects.

        :param bucket_name: Bucket name to list uploaded objects.
        :param prefix: String specifying objects returned must begin with.
        :param recursive: If yes, returns all incomplete objects for a specified prefix.
        :return: An generator of incomplete uploads in alphabetical order.
        """
        is_valid_bucket_name(bucket_name)

        # If someone explicitly set prefix to None convert it to empty string.
        if prefix is None:
            prefix = ''

        # Initialize query parameters.
        query = {
            'uploads': '',
            'prefix': prefix
        }

        if not recursive:
            query['delimiter'] = '/'

        key_marker, upload_id_marker = '', ''
        is_truncated = True
        while is_truncated:
            if key_marker:
                query['key-marker'] = key_marker
            if upload_id_marker:
                query['upload-id-marker'] = upload_id_marker

            response = self._url_open('GET',
                                      bucket_name=bucket_name,
                                      query=query)
            (uploads, is_truncated, key_marker,
             upload_id_marker) = parse_list_multipart_uploads(response.data,
                                                              bucket_name)
            for upload in uploads:
                if is_aggregate_size:
                    upload.size = self._get_total_multipart_upload_size(
                        upload.bucket_name,
                        upload.object_name,
                        upload.upload_id)
                yield upload

    def _get_total_multipart_upload_size(self, bucket_name, object_name,
                                         upload_id):
        """
        Get total multipart upload size.

        :param bucket_name: Bucket name to list parts for.
        :param object_name: Object name to list parts for.
        :param upload_id: Upload id of the previously uploaded object name.
        """
        return sum(
            [part.size for part in
             self._list_object_parts(bucket_name, object_name, upload_id)]
        )

    def _list_object_parts(self, bucket_name, object_name, upload_id):
        """
        List all parts.

        :param bucket_name: Bucket name to list parts for.
        :param object_name: Object name to list parts for.
        :param upload_id: Upload id of the previously uploaded object name.
        """
        is_valid_bucket_name(bucket_name)
        is_non_empty_string(object_name)
        is_non_empty_string(upload_id)

        query = {
            'uploadId': upload_id,
        }

        is_truncated = True
        part_number_marker = ''
        while is_truncated:
            if part_number_marker:
                query['part-number-marker'] = str(part_number_marker)

            response = self._url_open('GET',
                                      bucket_name=bucket_name,
                                      object_name=object_name,
                                      query=query)

            parts, is_truncated, part_number_marker = parse_list_parts(
                response.data,
                bucket_name=bucket_name,
                object_name=object_name,
                upload_id=upload_id
            )
            for part in parts:
                yield part

    def remove_incomplete_upload(self, bucket_name, object_name):
        """
        Remove all in-complete uploads for a given bucket_name and object_name.

        :param bucket_name: Bucket to drop incomplete uploads
        :param object_name: Name of object to remove incomplete uploads
        :return: None
        """
        is_valid_bucket_name(bucket_name)
        is_non_empty_string(object_name)

        recursive = True
        uploads = self._list_incomplete_uploads(bucket_name, object_name,
                                                recursive,
                                                is_aggregate_size=False)
        for upload in uploads:
            if object_name == upload.object_name:
                self._remove_incomplete_upload(bucket_name, object_name,
                                               upload.upload_id)

    def presigned_url(self, method,
                    bucket_name,
                    object_name,
                    expires=timedelta(days=7),
                    response_headers=None,
                    request_date=None):
        """
        Presigns a method on an object and provides a url

        Example:
            from datetime import timedelta

            presignedURL = presigned_url('GET',
                                         'bucket_name',
                                         'object_name',
                                         expires=timedelta(days=7))
            print(presignedURL)

        :param bucket_name: Bucket for the presigned url.
        :param object_name: Object for which presigned url is generated.
        :param expires: Optional expires argument to specify timedelta.
                        Defaults to 7days.
        :params response_headers: Optional response_headers argument to
                                  specify response fields like date, size,
                                  type of file, data about server, etc.
        :params request_date: Optional request_date argument to
                              specify a different request date. Default is
                              current date.
        :return: Presigned put object url.
        """
        is_valid_bucket_name(bucket_name)
        is_non_empty_string(object_name)

        if expires.total_seconds() < 1 or \
           expires.total_seconds() > _MAX_EXPIRY_TIME:
            raise InvalidArgumentError('Expires param valid values'
                                       ' are between 1 sec to'
                                       ' {0} secs'.format(_MAX_EXPIRY_TIME))

        region = self._get_bucket_region(bucket_name)
        url = get_target_url(self._endpoint_url,
                             bucket_name=bucket_name,
                             object_name=object_name,
                             bucket_region=region)

        return presign_v4(method, url,
                          self._access_key,
                          self._secret_key,
                          session_token=self._session_token,
                          region=region,
                          expires=int(expires.total_seconds()),
                          response_headers=response_headers,
                          request_date=request_date)

    def presigned_get_object(self, bucket_name, object_name,
                             expires=timedelta(days=7),
                             response_headers=None,
                             request_date=None):
        """
        Presigns a get object request and provides a url

        Example:

            from datetime import timedelta

            presignedURL = presigned_get_object('bucket_name',
                                                'object_name',
                                                timedelta(days=7))
            print(presignedURL)

        :param bucket_name: Bucket for the presigned url.
        :param object_name: Object for which presigned url is generated.
        :param expires: Optional expires argument to specify timedelta.
           Defaults to 7days.
        :params response_headers: Optional response_headers argument to
                                  specify response fields like date, size,
                                  type of file, data about server, etc.
        :params request_date: Optional request_date argument to
                              specify a different request date. Default is
                              current date.
        :return: Presigned url.
        """

        return self.presigned_url('GET',
                                  bucket_name,
                                  object_name,
                                  expires,
                                  response_headers=response_headers,
                                  request_date=request_date)

    def presigned_put_object(self, bucket_name, object_name,
                             expires=timedelta(days=7)):
        """
        Presigns a put object request and provides a url

        Example:
            from datetime import timedelta

            presignedURL = presigned_put_object('bucket_name',
                                                'object_name',
                                                timedelta(days=7))
            print(presignedURL)

        :param bucket_name: Bucket for the presigned url.
        :param object_name: Object for which presigned url is generated.
        :param expires: optional expires argument to specify timedelta.
           Defaults to 7days.
        :return: Presigned put object url.
        """

        return self.presigned_url('PUT',
                                  bucket_name,
                                  object_name,
                                  expires)

    def presigned_post_policy(self, post_policy):
        """
        Provides a POST form data that can be used for object uploads.

        Example:
            post_policy = PostPolicy()
            post_policy.set_bucket_name('bucket_name')
            post_policy.set_key_startswith('objectPrefix/')

            expires_date = datetime.utcnow()+timedelta(days=10)
            post_policy.set_expires(expires_date)

            print(presigned_post_policy(post_policy))

        :param post_policy: Post_Policy object.
        :return: PostPolicy form dictionary to be used in curl or HTML forms.
        """
        post_policy.is_valid()

        date = datetime.utcnow()
        iso8601_date = date.strftime("%Y%m%dT%H%M%SZ")
        region = self._get_bucket_region(post_policy.form_data['bucket'])
        credential_string = generate_credential_string(self._access_key,
                                                       date, region)

        policy = [
            ('eq', '$x-amz-date', iso8601_date),
            ('eq', '$x-amz-algorithm', _SIGN_V4_ALGORITHM),
            ('eq', '$x-amz-credential', credential_string),
        ]
        if self._session_token:
            policy.append(('eq', '$x-amz-security-token', self._session_token))

        post_policy_base64 = post_policy.base64(extras=policy)
        signature = post_presign_signature(date, region,
                                           self._secret_key,
                                           post_policy_base64)
        form_data = {
            'policy': post_policy_base64,
            'x-amz-algorithm': _SIGN_V4_ALGORITHM,
            'x-amz-credential': credential_string,
            'x-amz-date': iso8601_date,
            'x-amz-signature': signature,
        }
        if self._session_token:
            form_data['x-amz-security-token'] = self._session_token

        post_policy.form_data.update(form_data)
        url_str = get_target_url(self._endpoint_url,
                                 bucket_name=post_policy.form_data['bucket'],
                                 bucket_region=region)
        return (url_str, post_policy.form_data)

    # All private functions below.
    def _get_partial_object(self, bucket_name, object_name,
                            offset=0, length=0, request_headers=None, sse=None):
        """Retrieves an object from a bucket.

        Optionally takes an offset and length of data to retrieve.

        It returns a response object whose content is not
        pre-loaded. This means that the connection associated with the
        response needs to be released (for efficient re-use) after
        usage with `response.release_conn()`. Otherwise, the
        connection will linger until the object is garbage collected,
        when the connection is simply closed and not re-used.

        Examples:
            partial_object = minio.get_partial_object('foo', 'bar', 2, 4)

        :param bucket_name: Bucket to retrieve object from
        :param object_name: Name of object to retrieve
        :param offset: Optional offset to retrieve bytes from.
           Must be >= 0.
        :param length: Optional number of bytes to retrieve.
           Must be > 0.
        :param request_headers: Any additional metadata to be uploaded along
            with request.
        :return: :class:`urllib3.response.HTTPResponse` object.

        """
        is_valid_sse_c_object(sse=sse)
        is_valid_bucket_name(bucket_name)
        is_non_empty_string(object_name)

        headers = {}
        if request_headers:
            headers = request_headers

        if offset != 0 or length != 0:
            request_range = '{}-{}'.format(
                offset, "" if length == 0 else offset + length - 1
            )
            headers['Range'] = 'bytes=' + request_range

        if sse:
            headers.update(sse.marshal())

        return self._url_open('GET',
                               bucket_name=bucket_name,
                               object_name=object_name,
                               headers=headers,
                               preload_content=False)

    def _do_put_object(self, bucket_name, object_name, part_data,
                       part_size, upload_id='', part_number=0,
                       metadata=None, sse=None, progress=None):
        """
        Initiate a multipart PUT operation for a part number
        or single PUT object.

        :param bucket_name: Bucket name for the multipart request.
        :param object_name: Object name for the multipart request.
        :param part_metadata: Part-data and metadata for the multipart request.
        :param upload_id: Upload id of the multipart request [OPTIONAL].
        :param part_number: Part number of the data to be uploaded [OPTIONAL].
        :param metadata: Any additional metadata to be uploaded along
           with your object.
        :param progress: A progress object
        """
        is_valid_bucket_name(bucket_name)
        is_non_empty_string(object_name)

        # Accept only bytes - otherwise we need to know how to encode
        # the data to bytes before storing in the object.
        if not isinstance(part_data, bytes):
            raise ValueError('Input data must be bytes type')

        headers = {
            'Content-Length': part_size,
        }

        md5_base64 = ''
        sha256_hex = ''
        if self._is_ssl:
            md5_base64 = get_md5_base64digest(part_data)
            sha256_hex = _UNSIGNED_PAYLOAD
        else:
            sha256_hex = get_sha256_hexdigest(part_data)

        if md5_base64:
            headers['Content-Md5'] = md5_base64

        if metadata:
            headers.update(metadata)

        query = {}
        if part_number > 0 and upload_id:
            query = {
                'uploadId': upload_id,
                'partNumber': str(part_number),
            }
            # Encryption headers for multipart uploads should
            # be set only in the case of SSE-C.
            if sse and sse.type() == "SSE-C":
                headers.update(sse.marshal())
        elif sse:
            headers.update(sse.marshal())

        response = self._url_open(
            'PUT',
            bucket_name=bucket_name,
            object_name=object_name,
            query=query,
            headers=headers,
            body=io.BytesIO(part_data),
            content_sha256=sha256_hex
        )

        if progress:
            # Update the 'progress' object with uploaded 'part_size'.
            progress.update(part_size)
        return response.headers['etag'].replace('"', '')

    def _upload_part_routine(self, part_info):
        bucket_name, object_name, upload_id, \
                part_number, part_data, sse, progress = part_info
        # Initiate multipart put.
        etag = self._do_put_object(bucket_name, object_name, part_data,
                                   len(part_data), upload_id,
                                   part_number, sse=sse, progress=progress)

        return part_number, etag, len(part_data)

    def _stream_put_object(self, bucket_name, object_name,
                           data, content_size,
                           metadata=None, sse=None,
                           progress=None, part_size=MIN_PART_SIZE):
        """
        Streaming multipart upload operation.

        :param bucket_name: Bucket name of the multipart upload.
        :param object_name: Object name of the multipart upload.
        :param content_size: Total size of the content to be uploaded.
        :param content_type: Content type of of the multipart upload.
           Defaults to 'application/octet-stream'.
        :param metadata: Any additional metadata to be uploaded along
           with your object.
        :param progress: A progress object
        :param part_size: Multipart part size
        """
        is_valid_bucket_name(bucket_name)
        is_non_empty_string(object_name)
        if not callable(getattr(data, 'read')):
            raise ValueError(
                'Invalid input data does not implement a callable read() method')

        # get upload id.
        upload_id = self._new_multipart_upload(bucket_name, object_name,
                                               metadata, sse)

        # Initialize variables
        total_uploaded = 0
        uploaded_parts = {}

        # Calculate optimal part info.
        total_parts_count, part_size, last_part_size = optimal_part_info(
            content_size, part_size)

        # Instantiate a thread pool with 3 worker threads
        pool = ThreadPool(_PARALLEL_UPLOADERS)
        pool.start_parallel()

        # Generate new parts and upload <= current_part_size until
        # part_number reaches total_parts_count calculated for the
        # given size. Additionally part_manager() also provides
        # md5digest and sha256digest for the partitioned data.
        for part_number in range(1, total_parts_count + 1):
            current_part_size = (part_size if part_number < total_parts_count
                                 else last_part_size)

            part_data = read_full(data, current_part_size)
            pool.add_task(self._upload_part_routine, (bucket_name, object_name, upload_id,
                                                      part_number, part_data, sse, progress))

        try:
            upload_result = pool.result()
        except:
            # Any exception that occurs sends an abort on the
            # on-going multipart operation.
            self._remove_incomplete_upload(bucket_name,
                                           object_name,
                                           upload_id)
            raise

        # Update uploaded_parts with the part uploads result
        # and check total uploaded data.
        while not upload_result.empty():
            part_number, etag, total_read = upload_result.get()
            uploaded_parts[part_number] = UploadPart(bucket_name,
                                                     object_name,
                                                     upload_id,
                                                     part_number,
                                                     etag,
                                                     None,
                                                     total_read)

            total_uploaded += total_read

        if total_uploaded != content_size:
            msg = 'Data uploaded {0} is not equal input size ' \
                  '{1}'.format(total_uploaded, content_size)
            # cleanup incomplete upload upon incorrect upload
            # automatically
            self._remove_incomplete_upload(bucket_name,
                                           object_name,
                                           upload_id)
            raise InvalidSizeError(msg)

        # Complete all multipart transactions if possible.
        try:
            mpart_result = self._complete_multipart_upload(bucket_name,
                                                           object_name,
                                                           upload_id,
                                                           uploaded_parts)
        except:
            # Any exception that occurs sends an abort on the
            # on-going multipart operation.
            self._remove_incomplete_upload(bucket_name,
                                           object_name,
                                           upload_id)
            raise

        # Return etag here.
        return mpart_result.etag

    def _remove_incomplete_upload(self, bucket_name, object_name, upload_id):
        """
        Remove incomplete multipart request.

        :param bucket_name: Bucket name of the incomplete upload.
        :param object_name: Object name of incomplete upload.
        :param upload_id: Upload id of the incomplete upload.
        """

        # No reason to store successful response, for errors
        # relevant exceptions are thrown.
        self._url_open('DELETE', bucket_name=bucket_name,
                       object_name=object_name, query={'uploadId': upload_id},
                       headers={})

    def _new_multipart_upload(self, bucket_name, object_name,
                              metadata=None, sse=None):
        """
        Initialize new multipart upload request.

        :param bucket_name: Bucket name of the new multipart request.
        :param object_name: Object name of the new multipart request.
        :param metadata: Additional new metadata for the new object.
        :return: Returns an upload id.
        """
        is_valid_bucket_name(bucket_name)
        is_non_empty_string(object_name)

        headers = {}
        if metadata:
            headers.update(metadata)
        if sse:
            headers.update(sse.marshal())

        response = self._url_open('POST', bucket_name=bucket_name,
                                  object_name=object_name,
                                  query={'uploads': ''},
                                  headers=headers)

        return parse_new_multipart_upload(response.data)

    def _complete_multipart_upload(self, bucket_name, object_name,
                                   upload_id, uploaded_parts):
        """
        Complete an active multipart upload request.

        :param bucket_name: Bucket name of the multipart request.
        :param object_name: Object name of the multipart request.
        :param upload_id: Upload id of the active multipart request.
        :param uploaded_parts: Key, Value dictionary of uploaded parts.
        """
        is_valid_bucket_name(bucket_name)
        is_non_empty_string(object_name)
        is_non_empty_string(upload_id)

        # Order uploaded parts as required by S3 specification
        ordered_parts = []
        for part in sorted(uploaded_parts.keys()):
            ordered_parts.append(uploaded_parts[part])

        data = xml_marshal_complete_multipart_upload(ordered_parts)
        sha256_hex = get_sha256_hexdigest(data)
        md5_base64 = get_md5_base64digest(data)

        headers = {
            'Content-Length': len(data),
            'Content-Type': 'application/xml',
            'Content-Md5': md5_base64,
        }

        response = self._url_open('POST', bucket_name=bucket_name,
                                  object_name=object_name,
                                  query={'uploadId': upload_id},
                                  headers=headers, body=data,
                                  content_sha256=sha256_hex)

        return parse_multipart_upload_result(response.data)

    def _delete_bucket_region(self, bucket_name):
        """
        Delete a bucket from bucket region cache.

        :param bucket_name: Bucket name to be removed from cache.
        """

        # Handles if bucket doesn't exist as well.
        self._region_map.pop(bucket_name, None)

    def _set_bucket_region(self, bucket_name, region='us-east-1'):
        """
        Sets a bucket region into bucket region cache.

        :param bucket_name: Bucket name for which region is set.
        :param region: Region of the bucket name to set.
        """
        self._region_map[bucket_name] = region

    def _get_bucket_region(self, bucket_name):
        """
        Get region based on the bucket name.

        :param bucket_name: Bucket name for which region will be fetched.
        :return: Region of bucket name.
        """

        # Region set in constructor, return right here.
        if self._region:
            return self._region

        # get bucket location for Amazon S3.
        region = 'us-east-1'  # default to US standard.
        if bucket_name in self._region_map:
            region = self._region_map[bucket_name]
        else:
            region = self._get_bucket_location(bucket_name)
            self._region_map[bucket_name] = region

        # Success.
        return region

    def _get_bucket_location(self, bucket_name):
        """
        Get bucket location.

        :param bucket_name: Fetches location of the Bucket name.
        :return: location of bucket name is returned.
        """
        method = 'GET'
        url = self._endpoint_url + '/' + bucket_name + '?location='
        headers = {}
        # default for all requests.
        region = 'us-east-1'

        # Region is set override.
        if self._region:
            return self._region

        # For anonymous requests no need to get bucket location.
        if self._access_key is None or self._secret_key is None:
            return 'us-east-1'

        # Get signature headers if any.
        headers = sign_v4(method, url, region,
                          headers, self._access_key,
                          self._secret_key,
                          self._session_token,
                          None, datetime.utcnow())

        response = self._http.urlopen(method, url,
                                      body=None,
                                      headers=headers)

        if self._trace_output_stream:
            dump_http(method, url, headers, response,
                      self._trace_output_stream)

        if response.status != 200:
            raise ResponseError(response, method, bucket_name).get_exception()

        location = parse_location_constraint(response.data)
        # location is empty for 'US standard region'
        if not location:
            return 'us-east-1'
        # location can be 'EU' convert it to meaningful 'eu-west-1'
        if location == 'EU':
            return 'eu-west-1'
        return location

    def _url_open(self, method, bucket_name=None, object_name=None,
                  query=None, body=None, headers={}, content_sha256=None,
                  preload_content=True):
        """
        Open a url wrapper around signature version '4'
           and :meth:`urllib3.PoolManager.urlopen`
        """
        # HTTP headers are case insensitive filter out
        # all duplicate headers and pick one.
        fold_case_headers = FoldCaseDict()

        # Set user agent once before executing the request.
        fold_case_headers['User-Agent'] = self._user_agent
        for header in headers:
            fold_case_headers[header] = headers[header]

        # Get bucket region.
        region = self._get_bucket_region(bucket_name)

        # Construct target url.
        url = get_target_url(self._endpoint_url, bucket_name=bucket_name,
                             object_name=object_name, bucket_region=region,
                             query=query)

        # Get signature headers if any.
        headers = sign_v4(method, url, region,
                          fold_case_headers, self._access_key,
                          self._secret_key, self._session_token,
                          content_sha256, datetime.utcnow())

        response = self._http.urlopen(method, url,
                                      body=body,
                                      headers=headers,
                                      preload_content=preload_content)

        if self._trace_output_stream:
            dump_http(method, url, fold_case_headers, response,
                      self._trace_output_stream)

        if response.status != 200 and response.status != 204 \
           and response.status != 206:
            # Upon any response error invalidate the region cache
            # proactively for the bucket name.
            self._delete_bucket_region(bucket_name)

            # In case we did not preload_content, we need to release
            # the connection:
            if not preload_content:
                response.release_conn()

            supported_methods = [
                'HEAD',
                'GET',
                'POST',
                'PUT',
                'DELETE'
            ]

            if method in supported_methods:
                raise ResponseError(response,
                                    method,
                                    bucket_name,
                                    object_name).get_exception()
            else:
                raise ValueError('Unsupported method returned'
                                 ' error: {0}'.format(response.status))

        return response



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2015,2016 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from minio.compat import _is_py3
from minio.fold_case_dict import FoldCaseDict

if _is_py3:
    import http.client as httplib
else:
    import httplib

from nose.tools import eq_

class MockResponse(object):
    def __init__(self, method, url, headers, status_code, response_headers=None,
                 content=None):
        self.method = method
        self.url = url
        self.request_headers = FoldCaseDict()
        for header in headers:
            self.request_headers[header] = headers[header]
        self.status = status_code
        self.headers = response_headers
        self.data = content
        if content is None:
            self.reason = httplib.responses[status_code]

    # noinspection PyUnusedLocal
    def read(self, amt=1024):
        return self.data

    def mock_verify(self, method, url, headers):
        eq_(self.method, method)
        eq_(self.url, url)
        for header in headers:
            eq_(self.request_headers[header], headers[header])

    # noinspection PyUnusedLocal
    def stream(self, chunk_size=1, decode_unicode=False):
        if self.data is not None:
            return iter(bytearray(self.data, 'utf-8'))
        return iter([])

    # dummy release connection call.
    def release_conn(self):
        return

class MockConnection(object):
    def __init__(self):
        self.requests = []

    def mock_add_request(self, request):
        self.requests.append(request)

    # noinspection PyUnusedLocal
    def request(self, method, url, headers, redirect=False):
        return_request = self.requests.pop(0)
        return_request.mock_verify(method, url, headers)
        return return_request

    # noinspection PyRedeclaration,PyUnusedLocal,PyUnusedLocal
    def urlopen(self, method, url, headers, preload_content=False, body=None,
                redirect=False):
        return self.request(method, url, headers)



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage.
# Copyright (C) 2016 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Note: YOUR-ACCESSKEYID, YOUR-SECRETACCESSKEY, my-testfile, my-bucketname and
# my-objectname are dummy values, please replace them with original values.

from minio import Minio

client = Minio('play.min.io',
               access_key='Q3AM3UQ867SPQQA43P2F',
               secret_key='zuf+tfteSlswRu7BJ86wekitnifILbZam1KYY3TG')

# Put a file with default content-type.
events = client.listen_bucket_notification('my-bucket', 'my-prefix/',
                                           '.my-suffix',
                                           ['s3:ObjectCreated:*',
                                            's3:ObjectRemoved:*',
                                            's3:ObjectAccessed:*'])
for event in events:
    print(event)



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2015, 2016 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
minio.compat
~~~~~~~~~~~~

This module implements python 2.x and 3.x compatibility layer.

:copyright: (c) 2015, 2016 by MinIO, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

import sys

#: Python 2.x?
_is_py2 = (sys.version_info[0] == 2)

#: Python 3.x?
_is_py3 = (sys.version_info[0] == 3)

if _is_py2:
    from Queue import Queue
    queue = Queue

    from Queue import Empty
    queue_empty = Empty

    from urllib import quote
    _urlencode = quote

    from urllib import unquote
    urldecode = unquote

    import urlparse
    urlsplit = urlparse.urlsplit
    parse_qs = urlparse.parse_qs

    ## Create missing types.
    bytes = str

    ## Update better types.
    builtin_range = range
    range = xrange
    builtin_str = str
    str = unicode

    ## Add missing imports
    basestring = basestring
elif _is_py3:
    from queue import Queue
    queue = Queue

    from queue import Empty
    queue_empty = Empty

    from urllib.request import quote
    _urlencode = quote

    from urllib.request import unquote
    urldecode = unquote

    import urllib.parse
    urlsplit = urllib.parse.urlsplit
    parse_qs = urllib.parse.parse_qs

    ## Create types to compat with py2.
    builtin_range = range
    builtin_str = str

    ## Create missing types.
    basestring = (str, bytes)
    long = int

    ## Add missing imports
    bytes = bytes
    range = range
    str = str

numeric_types = (int, long, float)

def urlencode(resource):
    """
    This implementation of urlencode supports all unicode characters

    :param: resource: Resource value to be url encoded.
    """
    if isinstance(resource, str):
        return _urlencode(resource.encode('utf-8'))

    return _urlencode(resource)

def queryencode(query):
    """
    This implementation of queryencode supports all unicode characters

    :param: query: Query value to be url encoded.
    """
    return urlencode(query).replace('/', '%2F')



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C) 2015 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from nose.tools import raises, eq_
from unittest import TestCase

from minio import Minio
from minio.api import _DEFAULT_USER_AGENT
from minio.error import InvalidBucketError, ResponseError

from .minio_mocks import MockResponse, MockConnection

class BucketExists(TestCase):
    @raises(TypeError)
    def test_bucket_is_string(self):
        client = Minio('localhost:9000')
        client.bucket_exists(1234)

    @raises(InvalidBucketError)
    def test_bucket_is_not_empty_string(self):
        client = Minio('localhost:9000')
        client.bucket_exists('  \t \n  ')

    @raises(InvalidBucketError)
    def test_bucket_exists_invalid_name(self):
        client = Minio('localhost:9000')
        client.bucket_exists('ABCD')

    @mock.patch('urllib3.PoolManager')
    @raises(ResponseError)
    def test_bucket_exists_works(self, mock_connection):
        mock_server = MockConnection()
        mock_connection.return_value = mock_server
        mock_server.mock_add_request(MockResponse('HEAD',
                                                  'https://localhost:9000/hello/',
                                                  {'User-Agent': _DEFAULT_USER_AGENT},
                                                  400))
        client = Minio('localhost:9000')
        result = client.bucket_exists('hello')

    @mock.patch('urllib3.PoolManager')
    def test_bucket_exists_works(self, mock_connection):
        mock_server = MockConnection()
        mock_connection.return_value = mock_server
        mock_server.mock_add_request(MockResponse('HEAD',
                                                  'https://localhost:9000/hello/',
                                                  {'User-Agent': _DEFAULT_USER_AGENT},
                                                  200))
        client = Minio('localhost:9000')
        result = client.bucket_exists('hello')
        eq_(True, result)
        mock_server.mock_add_request(MockResponse('HEAD',
                                                  'https://localhost:9000/goodbye/',
                                                  {'User-Agent': _DEFAULT_USER_AGENT},
                                                  404))
        false_result = client.bucket_exists('goodbye')
        eq_(False, false_result)



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C) 2016 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import itertools

from unittest import TestCase
from nose.tools import raises

from minio import Minio
from minio.api import _DEFAULT_USER_AGENT
from minio.error import InvalidBucketError

from .minio_mocks import MockResponse, MockConnection

class RemoveObjectsTest(TestCase):
    @raises(TypeError)
    def test_object_is_non_string_iterable_1(self):
        client = Minio('localhost:9000')
        for err in client.remove_objects('hello', 1234):
            print(err)

    @raises(TypeError)
    def test_object_is_non_string_iterable_2(self):
        client = Minio('localhost:9000')
        for err in client.remove_objects('hello', u'abc'):
            print(err)

    @raises(TypeError)
    def test_object_is_non_string_iterable_3(self):
        client = Minio('localhost:9000')
        for err in client.remove_objects('hello', b'abc'):
            print(err)

    @raises(InvalidBucketError)
    def test_bucket_invalid_name(self):
        client = Minio('localhost:9000')
        for err in client.remove_objects('ABCD', 'world'):
            print(err)

    @mock.patch('urllib3.PoolManager')
    def test_object_is_list(self, mock_connection):
        mock_server = MockConnection()
        mock_connection.return_value = mock_server
        mock_server.mock_add_request(
            MockResponse('POST',
                         'https://localhost:9000/hello/?delete=',
                         {'Content-Length': 95,
                          'User-Agent': _DEFAULT_USER_AGENT,
                          'Content-Md5': u'5Tg5SmU9Or43L4+iIyfPrQ=='}, 200,
                         content='<Delete/>')
        )
        client = Minio('localhost:9000')
        for err in client.remove_objects('hello', ["Ab", "c"]):
            print(err)

    @mock.patch('urllib3.PoolManager')
    def test_object_is_tuple(self, mock_connection):
        mock_server = MockConnection()
        mock_connection.return_value = mock_server
        mock_server.mock_add_request(
            MockResponse('POST',
                         'https://localhost:9000/hello/?delete=',
                         {'Content-Length': 95,
                          'User-Agent': _DEFAULT_USER_AGENT,
                          'Content-Md5': u'5Tg5SmU9Or43L4+iIyfPrQ=='}, 200,
                         content='<Delete/>')
        )
        client = Minio('localhost:9000')
        for err in client.remove_objects('hello', ('Ab', 'c')):
            print(err)

    @mock.patch('urllib3.PoolManager')
    def test_object_is_iterator(self, mock_connection):
        mock_server = MockConnection()
        mock_connection.return_value = mock_server
        mock_server.mock_add_request(
            MockResponse('POST',
                         'https://localhost:9000/hello/?delete=',
                         {'Content-Length': 95,
                          'User-Agent': _DEFAULT_USER_AGENT,
                          'Content-Md5': u'5Tg5SmU9Or43L4+iIyfPrQ=='}, 200,
                         content='<Delete/>')
        )
        client = Minio('localhost:9000')
        it = itertools.chain(('Ab', 'c'))
        for err in client.remove_objects('hello', it):
            print(err)



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2019 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
minio.select.errors
~~~~~~~~~~~~~~~

This module implements the error classes for SelectObject responses.

:copyright: (c) 2019 by MinIO, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

class SelectMessageError(Exception):
    '''
    Raised in case of message type 'error'
    '''

class SelectCRCValidationError(Exception):
    '''
    Raised in case of CRC mismatch
    '''



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C)
#  2019 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from minio import Minio
from minio.error import ResponseError
from minio.select.errors import SelectCRCValidationError, SelectMessageError
from minio.select.options import (SelectObjectOptions, CSVInput,
                                  JSONInput, RequestProgress,
                                  ParquetInput, InputSerialization,
                                  OutputSerialization, CSVOutput,
                                  JsonOutput)

client = Minio('s3.amazonaws.com',
               access_key='YOUR-ACCESSKEY',
               secret_key='YOUR-SECRETKEY')

options = SelectObjectOptions(
    expression="select * from s3object",
    input_serialization=InputSerialization(
        compression_type="NONE",
        csv=CSVInput(FileHeaderInfo="USE",
                     RecordDelimiter="\n",
                     FieldDelimiter=",",
                     QuoteCharacter='"',
                     QuoteEscapeCharacter='"',
                     Comments="#",
                     AllowQuotedRecordDelimiter="FALSE",
                     ),
        # If input is JSON
        # json=JSONInput(Type="DOCUMENT",)
        ),

    output_serialization=OutputSerialization(
        csv=CSVOutput(QuoteFields="ASNEEDED",
                      RecordDelimiter="\n",
                      FieldDelimiter=",",
                      QuoteCharacter='"',
                      QuoteEscapeCharacter='"',)

        # json = JsonOutput(
        #     RecordDelimiter="\n",
        #     )
        ),
    request_progress=RequestProgress(
        enabled="False"
        )
    )

try:
    data = client.select_object_content('your-bucket', 'your-object', options)

    # Get the records
    with open('my-record-file', 'w') as record_data:
        for d in data.stream(10*1024):
            record_data.write(d)

    # Get the stats
    print(data.stats())

except SelectMessageError as err:
    print(err)

except SelectCRCValidationError as err:
    print(err)

except ResponseError as err:
    print(err)



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2019 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
minio.select.reader
~~~~~~~~~~~~~~~

This module implements the reader for SelectObject response body.

:copyright: (c) 2019 by MinIO, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

from __future__ import absolute_import

import io
import sys

from binascii import crc32
from xml.etree import cElementTree
from xml.etree.cElementTree import ParseError

from .helpers import (EVENT_RECORDS, EVENT_PROGRESS,
                      EVENT_STATS, EVENT_CONT,
                      EVENT, EVENT_CONTENT_TYPE,
                      EVENT_END, ERROR)

from .helpers import (validate_crc, calculate_crc, byte_int)
from .errors import (SelectMessageError, SelectCRCValidationError)

def _extract_header(header_bytes):
    """
    populates the header map after reading the header in bytes
    """
    header_map = {}
    header_byte_parsed = 0
    # While loop ends when all the headers present are read
    # header contains multipe headers
    while header_byte_parsed < len(header_bytes):
        header_name_byte_length = byte_int(header_bytes[header_byte_parsed:header_byte_parsed+1])
        header_byte_parsed += 1
        header_name = \
            header_bytes[header_byte_parsed:
                         header_byte_parsed+header_name_byte_length]
        header_byte_parsed += header_name_byte_length
        # Header Value Type is of 1 bytes and is skipped
        header_byte_parsed += 1
        value_string_byte_length = \
            byte_int(header_bytes[header_byte_parsed:
                                  header_byte_parsed+2])
        header_byte_parsed += 2
        header_value = \
            header_bytes[header_byte_parsed:
                         header_byte_parsed+value_string_byte_length]
        header_byte_parsed += value_string_byte_length
        header_map[header_name.decode("utf-8").lstrip(":")] = \
            header_value.decode("utf-8").lstrip(":")
    return header_map

def _parse_stats(stats):
    """
    Parses stats XML and populates the stat dict.
    """
    stat = {}
    for attribute in cElementTree.fromstring(stats):
        if attribute.tag == 'BytesScanned':
            stat['BytesScanned'] = attribute.text
        elif attribute.tag == 'BytesProcessed':
            stat['BytesProcessed'] = attribute.text
        elif attribute.tag == 'BytesReturned':
            stat['BytesReturned'] = attribute.text

    return stat

class SelectObjectReader(object):
    """
    SelectObjectReader returns a Reader that upon read
    returns queried data, but stops when the response ends.
    LimitedRandomReader is compatible with BufferedIOBase.
    """
    def __init__(self, response):
        self.response = response
        self.remaining_bytes = bytes()
        self.stat = {}
        self.prog = {}

    def readable(self):
        return True

    def writeable(self):
        return False

    def close(self):
        self.response.close()

    def stats(self):
        return self.stat

    def progress(self):
        return self.prog

    def __extract_message(self):
        """
        Process the response sent from server.
        https://docs.aws.amazon.com/AmazonS3/latest/API/RESTObjectSELECTContent.html
        """

        crc_bytes = io.BytesIO()
        total_bytes_len = self.response.read(4)
        if len(total_bytes_len) == 0:
            return {}

        total_length = byte_int(total_bytes_len)
        header_bytes_len = self.response.read(4)
        if len(header_bytes_len) == 0:
            return {}

        header_len = byte_int(header_bytes_len)

        crc_bytes.write(total_bytes_len)
        crc_bytes.write(header_bytes_len)

        prelude_bytes_crc = self.response.read(4)
        if not validate_crc(crc_bytes.getvalue(), prelude_bytes_crc):
            raise SelectCRCValidationError(
                {"Checksum Mismatch, PreludeCRC of " +
                 str(calculate_crc(crc_bytes.getvalue())) +
                 " does not equal expected CRC of " +
                 str(byte_int(prelude_bytes_crc))})

        crc_bytes.write(prelude_bytes_crc)

        header_bytes = self.response.read(header_len)
        if len(header_bytes) == 0:
            raise SelectMessageError(
                "Premature truncation of select message header"+
                ", server is sending corrupt message?")

        crc_bytes.write(header_bytes)

        header_map = _extract_header(header_bytes)
        payload_length = total_length - header_len - int(16)
        payload_bytes = b''
        event_type = header_map["event-type"]
        if header_map["message-type"] == ERROR:
            raise SelectMessageError(
                header_map["error-code"] + ":\"" + \
                header_map["error-message"] + "\"")
        elif header_map["message-type"] == EVENT:
            if event_type == EVENT_END:
                pass
            elif event_type == EVENT_CONT:
                pass
            elif event_type == EVENT_STATS:
                content_type = header_map["content-type"]
                if content_type != EVENT_CONTENT_TYPE:
                    raise SelectMessageError(
                        "Unrecognized content-type {0}".format(content_type))
                else:
                    payload_bytes = self.response.read(payload_length)
                    self.stat = _parse_stats(payload_bytes)

            elif event_type == EVENT_RECORDS:
                payload_bytes = self.response.read(payload_length)
        else:
            raise SelectMessageError(
                "Unrecognized message-type {0}".format(header_map["message-type"])
            )

        crc_bytes.write(payload_bytes)

        message_crc = self.response.read(4)
        if len(message_crc) == 0:
            return {}

        if not validate_crc(crc_bytes.getvalue(),
                            message_crc):
            raise SelectCRCValidationError(
                {"Checksum Mismatch, MessageCRC of " +
                 str(calculate_crc(crc_bytes.getvalue())) +
                 " does not equal expected CRC of " +
                 str(byte_int(message_crc))})

        message = {event_type: payload_bytes}
        return message

    def stream(self, num_bytes=32*1024):
        """
        extract each record from the response body ... and buffer it.
        send only up to requested bytes such as message[:num_bytes]
        rest is buffered and added to the next iteration.

        caller should call self.close() to close the stream.
        """
        while not self.response.isclosed():
            if len(self.remaining_bytes) == 0:
                message = self.__extract_message()
                if EVENT_RECORDS in message:
                    self.remaining_bytes = message.get(EVENT_RECORDS, b'')
                else:
                    # For all other events continue
                    continue

            result = self.remaining_bytes
            if num_bytes < len(self.remaining_bytes):
                result = self.remaining_bytes[:num_bytes]
            self.remaining_bytes = self.remaining_bytes[len(result):]

            if result == b'':
                break
            if sys.version_info.major == 3:
                yield result.decode('utf-8', errors='ignore')
            else:
                # Python 2.x needs explicit conversion.
                yield result.decode('utf-8', errors='ignore').encode('utf-8')



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C) 2015 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

def generate_error(code, message, request_id, host_id,
                   resource, bucket_name, object_name):
    return '''
    <Error>
      <Code>{0}</Code>
      <Message>{1}</Message>
      <RequestId>{2}</RequestId>
      <HostId>{3}</HostId>
      <Resource>{4}</Resource>
      <BucketName>{5}</BucketName>
      <Key>{6}</Key>
    </Error>
    '''.format(code, message, request_id, host_id,
               resource, bucket_name, object_name)



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C) 2015 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from nose.tools import eq_, timed
from unittest import TestCase

from minio import Minio
from minio.api import _DEFAULT_USER_AGENT

from .minio_mocks import MockResponse, MockConnection

class ListObjectsV2Test(TestCase):
    @mock.patch('urllib3.PoolManager')
    def test_empty_list_objects_works(self, mock_connection):
        mock_data = '''<?xml version="1.0" encoding="UTF-8"?>
            <ListBucketResult xmlns="http://s3.amazonaws.com/doc/2006-03-01/">
              <Name>bucket</Name>
              <Prefix></Prefix>
              <KeyCount>0</KeyCount>
              <MaxKeys>1000</MaxKeys>
              <Delimiter></Delimiter>
              <IsTruncated>false</IsTruncated>
            </ListBucketResult>
        '''
        mock_server = MockConnection()
        mock_connection.return_value = mock_server
        mock_server.mock_add_request(MockResponse('GET',
                                                  'https://localhost:9000/bucket/?list-type=2&prefix=&start-after=',
                                                  {'User-Agent': _DEFAULT_USER_AGENT}, 200, content=mock_data))
        client = Minio('localhost:9000')
        object_iter = client.list_objects_v2('bucket', recursive=True)
        objects = []
        for obj in object_iter:
            objects.append(obj)
        eq_(0, len(objects))

    @timed(1)
    @mock.patch('urllib3.PoolManager')
    def test_list_objects_works(self, mock_connection):
        mock_data = '''<?xml version="1.0" encoding="UTF-8"?>
            <ListBucketResult xmlns="http://s3.amazonaws.com/doc/2006-03-01/">
              <Name>bucket</Name>
              <Prefix></Prefix>
              <KeyCount>2</KeyCount>
              <MaxKeys>1000</MaxKeys>
              <IsTruncated>false</IsTruncated>
              <Contents>
                <Key>6/f/9/6f9898076bb08572403f95dbb86c5b9c85e1e1b3</Key>
                <LastModified>2016-11-27T07:55:53.000Z</LastModified>
                <ETag>&quot;5d5512301b6b6e247b8aec334b2cf7ea&quot;</ETag>
                <Size>493</Size>
                <StorageClass>REDUCED_REDUNDANCY</StorageClass>
              </Contents>
              <Contents>
                <Key>b/d/7/bd7f6410cced55228902d881c2954ebc826d7464</Key>
                <LastModified>2016-11-27T07:10:27.000Z</LastModified>
                <ETag>&quot;f00483d523ffc8b7f2883ae896769d85&quot;</ETag>
                <Size>493</Size>
                <StorageClass>REDUCED_REDUNDANCY</StorageClass>
              </Contents>
            </ListBucketResult>
        '''
        mock_server = MockConnection()
        mock_connection.return_value = mock_server
        mock_server.mock_add_request(
            MockResponse(
                'GET',
                'https://localhost:9000/bucket/?delimiter=%2F&list-type=2&prefix=&start-after=',
                {'User-Agent': _DEFAULT_USER_AGENT}, 200,
                content=mock_data
            )
        )
        client = Minio('localhost:9000')
        objects_iter = client.list_objects_v2('bucket')
        objects = []
        for obj in objects_iter:
            objects.append(obj)

        eq_(2, len(objects))



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2019 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
minio.select.options
~~~~~~~~~~~~~~~

This module implements the SelectOption definition for SelectObject API.

:copyright: (c) 2019 by MinIO, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

from .helpers import (SQL)

class CSVInput:
    """
    CSVInput: Input Format as CSV.
    """
    def __init__(self, FileHeaderInfo=None, RecordDelimiter="\n",
                 FieldDelimiter=",", QuoteCharacter='"',
                 QuoteEscapeCharacter='"', Comments="#",
                 AllowQuotedRecordDelimiter=False):
        self.FileHeaderInfo = FileHeaderInfo
        self.RecordDelimiter = RecordDelimiter
        self.FieldDelimiter = FieldDelimiter
        self.QuoteCharacter = QuoteCharacter
        self.QuoteEscapeCharacter = QuoteEscapeCharacter
        self.Comments = Comments
        self.AllowQuotedRecordDelimiter = AllowQuotedRecordDelimiter

class JSONInput:
    """
    JSONInput: Input format as JSON.
    """
    def __init__(self, Type=None):
        self.Type = Type


class ParquetInput:
    """
    ParquetInput: Input format as Parquet
    """


class InputSerialization:
    """
    InputSerialization: nput Format.
    """
    def __init__(self, compression_type="NONE", csv=None, json=None, par=None):
        self.compression_type = compression_type
        self.csv_input = csv
        self.json_input = json
        self.parquet_input = par


class CSVOutput:
    """
    CSVOutput: Output as CSV.

    """
    def __init__(self, QuoteFields="ASNEEDED", RecordDelimiter="\n",
                 FieldDelimiter=",", QuoteCharacter='"',
                 QuoteEscapeCharacter='"'):
        self.QuoteFields = QuoteFields
        self.RecordDelimiter = RecordDelimiter
        self.FieldDelimiter = FieldDelimiter
        self.QuoteCharacter = QuoteCharacter
        self.QuoteEscapeCharacter = QuoteEscapeCharacter


class JsonOutput:
    """
    JsonOutput- Output as JSON.
    """
    def __init__(self, RecordDelimiter="\n"):
        self.RecordDelimiter = RecordDelimiter


class OutputSerialization:
    """
    OutputSerialization: Output Format.
    """
    def __init__(self,  csv=None, json=None):
        self.csv_output = csv
        self.json_output = json


class RequestProgress:
    """
    RequestProgress: Sends progress message.
    """
    def __init__(self, enabled=False):
        self.enabled = enabled


class SelectObjectOptions:
    """
    SelectObjectOptions: Options for select object
    """
    expression_type = SQL

    def __init__(self, expression, input_serialization,
                 output_serialization, request_progress):
        self.expression = expression
        self.in_ser = input_serialization
        self.out_ser = output_serialization
        self.req_progress = request_progress



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C) 2018 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Note: YOUR-ACCESSKEYID, YOUR-SECRETACCESSKEY and my-bucketname are
# dummy values, please replace them with original values.

from io import BytesIO
from minio.api import Minio
from minio.sse import SSE_S3


AWSAccessKeyId = 'YOUR-ACCESSKEYID'
AWSSecretKey = 'YOUR-SECRETACCESSKEY'

STORAGE_ENDPOINT = 's3.amazonaws.com'
STORAGE_BUCKET = 'test-encryption-bucket'


def main():
   
    minio = Minio(STORAGE_ENDPOINT, access_key=AWSAccessKeyId, secret_key=AWSSecretKey)

    content = BytesIO(b'Hello again')
    
    #Create an SSE_S3 object
    sse_s3_obj = SSE_S3()

    # Put object with from SSE_S3 object which encrypt object in S3 with provided key
    minio.put_object(STORAGE_BUCKET, 'test_crypt.txt', content, content.getbuffer().nbytes, sse=sse_s3_obj)

    # Get decrypted object with same headers
    obj = minio.get_object(STORAGE_BUCKET, 'test_crypt.txt')
   
    print(obj.read())

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage,
# (C) 2015, 2016, 2017 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase

from nose.tools import raises, eq_

from minio import Minio
from minio import __version__ as minio_version
from minio.api import _DEFAULT_USER_AGENT
from minio.error import InvalidEndpointError, InvalidBucketError
from minio.helpers import (get_target_url, is_valid_bucket_name,
                           get_s3_region_from_endpoint)

class ValidBucketName(TestCase):
    @raises(InvalidBucketError)
    def test_bucket_name(self):
        is_valid_bucket_name('bucketName')

    @raises(InvalidBucketError)
    def test_bucket_name_invalid_characters(self):
        is_valid_bucket_name('$$$bcuket')

    @raises(InvalidBucketError)
    def test_bucket_name_length(self):
        is_valid_bucket_name('dd')

    @raises(InvalidBucketError)
    def test_bucket_name_periods(self):
        is_valid_bucket_name('dd..mybucket')

    @raises(InvalidBucketError)
    def test_bucket_name_begins_period(self):
        is_valid_bucket_name('.ddmybucket')

class GetURLTests(TestCase):
    def test_get_target_url_works(self):
        url = 'http://localhost:9000'
        eq_(get_target_url(url, 'bucket-name'),
            'http://localhost:9000/bucket-name/')
        eq_(get_target_url(url, 'bucket-name', 'objectName'),
            'http://localhost:9000/bucket-name/objectName')
        eq_(get_target_url(url, 'bucket-name', 'objectName', None),
            'http://localhost:9000/bucket-name/objectName')
        eq_(get_target_url(url, 'bucket-name', 'objectName', 'us-east-1',
                           {'foo': 'bar'}),
            'http://localhost:9000/bucket-name/objectName?foo=bar')
        eq_(get_target_url(url, 'bucket-name', 'objectName', 'us-east-1',
                           {'foo': 'bar',
                            'b': 'c',
                            'a': 'b'}),
            'http://localhost:9000/bucket-name/objectName?a=b&b=c&foo=bar')
        # S3 urls.
        s3_url = 'https://s3.amazonaws.com'
        eq_(get_target_url(s3_url), 'https://s3.amazonaws.com/')
        eq_(get_target_url(s3_url, 'my.bucket.name'),
            'https://s3.amazonaws.com/my.bucket.name/')
        eq_(get_target_url(s3_url,
                           'bucket-name',
                           'objectName',
                           'us-west-2', None),
            'https://bucket-name.s3-us-west-2.amazonaws.com/objectName')

    @raises(TypeError)
    def test_minio_requires_string(self):
        Minio(10)

    @raises(InvalidEndpointError)
    def test_minio_requires_hostname(self):
        Minio('http://')


class UserAgentTests(TestCase):
    def test_default_user_agent(self):
        client = Minio('localhost')
        eq_(client._user_agent, _DEFAULT_USER_AGENT)

    def test_set_app_info(self):
        client = Minio('localhost')
        expected_user_agent = _DEFAULT_USER_AGENT + ' hello/' + minio_version
        client.set_app_info('hello', minio_version)
        eq_(client._user_agent, expected_user_agent)

    @raises(ValueError)
    def test_set_app_info_requires_non_empty_name(self):
        client = Minio('localhost:9000')
        client.set_app_info('', minio_version)

    @raises(ValueError)
    def test_set_app_info_requires_non_empty_version(self):
        client = Minio('localhost:9000')
        client.set_app_info('hello', '')

class GetRegionTests(TestCase):
    def test_region_none(self):
        region = get_s3_region_from_endpoint('localhost')
        eq_(region, None)

    def test_region_us_west(self):
        region = get_s3_region_from_endpoint('s3-us-west-1.amazonaws.com')
        eq_(region, 'us-west-1')

    def test_region_with_dot(self):
        region = get_s3_region_from_endpoint('s3.us-west-1.amazonaws.com')
        eq_(region, 'us-west-1')

    def test_region_with_dualstack(self):
        region = get_s3_region_from_endpoint('s3.dualstack.us-west-1.amazonaws.com')
        eq_(region, 'us-west-1')

    def test_region_us_east(self):
        region = get_s3_region_from_endpoint('s3.amazonaws.com')
        eq_(region, None)

    @raises(TypeError)
    def test_invalid_value(self):
        region = get_s3_region_from_endpoint(None)



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2015, 2016, 2017, 2018, 2019 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
minio.xml_marshal
~~~~~~~~~~~~~~~

This module contains the simple wrappers for XML marshaller's.

:copyright: (c) 2015 by MinIO, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

from __future__ import absolute_import
import io

from xml.etree import ElementTree as s3_xml

_S3_NAMESPACE = 'http://s3.amazonaws.com/doc/2006-03-01/'


def xml_marshal_bucket_constraint(region):
    """
    Marshal's bucket constraint based on *region*.

    :param region: Region name of a given bucket.
    :return: Marshalled XML data.
    """
    root = s3_xml.Element('CreateBucketConfiguration', {'xmlns': _S3_NAMESPACE})
    location_constraint = s3_xml.SubElement(root, 'LocationConstraint')
    location_constraint.text = region
    data = io.BytesIO()
    s3_xml.ElementTree(root).write(data, encoding=None, xml_declaration=False)
    return data.getvalue()


def xml_marshal_select(opts):
    root = s3_xml.Element('SelectObjectContentRequest')
    expression = s3_xml.SubElement(root, 'Expression')
    expression.text = opts.expression
    expression_type = s3_xml.SubElement(root, 'ExpressionType')
    expression_type.text = 'SQL'
    input_serialization = s3_xml.SubElement(root, 'InputSerialization')

    if opts.in_ser.csv_input is not None:
        compression_type = s3_xml.SubElement(input_serialization, 'CompressionType')
        compression_type.text = opts.in_ser.compression_type
        csv = s3_xml.SubElement(input_serialization, 'CSV')
        file_header_info = s3_xml.SubElement(csv, 'FileHeaderInfo')
        file_header_info.text = opts.in_ser.csv_input.FileHeaderInfo
        record_delimiter = s3_xml.SubElement(csv, 'RecordDelimiter')
        record_delimiter.text = opts.in_ser.csv_input.RecordDelimiter
        field_delimiter = s3_xml.SubElement(csv, 'FieldDelimiter')
        field_delimiter.text = opts.in_ser.csv_input.FieldDelimiter
        quote_character = s3_xml.SubElement(csv, 'QuoteCharacter')
        quote_character.text = opts.in_ser.csv_input.QuoteCharacter
        quote_escape_character = s3_xml.SubElement(csv, 'QuoteEscapeCharacter')
        quote_escape_character.text = opts.in_ser.csv_input.QuoteEscapeCharacter
        comments = s3_xml.SubElement(csv, 'Comments')
        comments.text = opts.in_ser.csv_input.Comments
        allow_quoted_record_delimiter = s3_xml.SubElement(csv, 'AllowQuotedRecordDelimiter')
        allow_quoted_record_delimiter.text = opts.in_ser.csv_input.AllowQuotedRecordDelimiter.lower()

    if opts.in_ser.json_input is not None:
        compression_type = s3_xml.SubElement(input_serialization, 'CompressionType')
        compression_type.text = opts.in_ser.compression_type
        json = s3_xml.SubElement(input_serialization, 'JSON')
        type_input = s3_xml.SubElement(json, 'Type')
        type_input.text = opts.in_ser.json_input.Type

    if opts.in_ser.parquet_input is not None:
        compression_type = s3_xml.SubElement(input_serialization, 'CompressionType')
        compression_type.text = opts.in_ser.compression_type
        s3_xml.SubElement(input_serialization, 'Parquet')

    output_serialization = s3_xml.SubElement(root, 'OutputSerialization')
    if opts.out_ser.csv_output is not None:
        csv = s3_xml.SubElement(output_serialization, 'CSV')
        quote_field = s3_xml.SubElement(csv, 'QuoteFields')
        quote_field.text = opts.out_ser.csv_output.QuoteFields
        record_delimiter = s3_xml.SubElement(csv, 'RecordDelimiter')
        record_delimiter.text = opts.out_ser.csv_output.RecordDelimiter
        field_delimiter = s3_xml.SubElement(csv, 'FieldDelimiter')
        field_delimiter.text = opts.out_ser.csv_output.FieldDelimiter
        quote_character = s3_xml.SubElement(csv, 'QuoteCharacter')
        quote_character.text = opts.out_ser.csv_output.QuoteCharacter
        quote_escape_character = s3_xml.SubElement(csv, 'QuoteEscapeCharacter')
        quote_escape_character.text = opts.out_ser.csv_output.QuoteEscapeCharacter

    if opts.out_ser.json_output is not None:
        json = s3_xml.SubElement(output_serialization, 'JSON')
        record_delimiter = s3_xml.SubElement(json, 'RecordDelimiter')
        record_delimiter.text = opts.out_ser.json_output.RecordDelimiter

    request_progress = s3_xml.SubElement(root, 'RequestProgress')
    enabled = s3_xml.SubElement(request_progress, 'Enabled')
    enabled.text = opts.req_progress.enabled.lower()

    data = io.BytesIO()
    s3_xml.ElementTree(root).write(data, encoding=None, xml_declaration=False)
    return data.getvalue()


def xml_marshal_complete_multipart_upload(uploaded_parts):
    """
    Marshal's complete multipart upload request based on *uploaded_parts*.

    :param uploaded_parts: List of all uploaded parts, ordered by part number.
    :return: Marshalled XML data.
    """
    root = s3_xml.Element('CompleteMultipartUpload', {'xmlns': _S3_NAMESPACE})
    for uploaded_part in uploaded_parts:
        part_number = uploaded_part.part_number
        part = s3_xml.SubElement(root, 'Part')
        part_num = s3_xml.SubElement(part, 'PartNumber')
        part_num.text = str(part_number)
        etag = s3_xml.SubElement(part, 'ETag')
        etag.text = '"' + uploaded_part.etag + '"'
        data = io.BytesIO()
        s3_xml.ElementTree(root).write(data, encoding=None,
                                       xml_declaration=False)
    return data.getvalue()


def xml_marshal_bucket_notifications(notifications):
    """
    Marshals the notifications structure for sending to S3 compatible storage

    :param notifications: Dictionary with following structure:

    {
        'TopicConfigurations': [
            {
                'Id': 'string',
                'Arn': 'string',
                'Events': [
                    's3:ReducedRedundancyLostObject'|'s3:ObjectCreated:*'|'s3:ObjectCreated:Put'|'s3:ObjectCreated:Post'|'s3:ObjectCreated:Copy'|'s3:ObjectCreated:CompleteMultipartUpload'|'s3:ObjectRemoved:*'|'s3:ObjectRemoved:Delete'|'s3:ObjectRemoved:DeleteMarkerCreated',
                ],
                'Filter': {
                    'Key': {
                        'FilterRules': [
                            {
                                'Name': 'prefix'|'suffix',
                                'Value': 'string'
                            },
                        ]
                    }
                }
            },
        ],
        'QueueConfigurations': [
            {
                'Id': 'string',
                'Arn': 'string',
                'Events': [
                    's3:ReducedRedundancyLostObject'|'s3:ObjectCreated:*'|'s3:ObjectCreated:Put'|'s3:ObjectCreated:Post'|'s3:ObjectCreated:Copy'|'s3:ObjectCreated:CompleteMultipartUpload'|'s3:ObjectRemoved:*'|'s3:ObjectRemoved:Delete'|'s3:ObjectRemoved:DeleteMarkerCreated',
                ],
                'Filter': {
                    'Key': {
                        'FilterRules': [
                            {
                                'Name': 'prefix'|'suffix',
                                'Value': 'string'
                            },
                        ]
                    }
                }
            },
        ],
        'CloudFunctionConfigurations': [
            {
                'Id': 'string',
                'Arn': 'string',
                'Events': [
                    's3:ReducedRedundancyLostObject'|'s3:ObjectCreated:*'|'s3:ObjectCreated:Put'|'s3:ObjectCreated:Post'|'s3:ObjectCreated:Copy'|'s3:ObjectCreated:CompleteMultipartUpload'|'s3:ObjectRemoved:*'|'s3:ObjectRemoved:Delete'|'s3:ObjectRemoved:DeleteMarkerCreated',
                ],
                'Filter': {
                    'Key': {
                        'FilterRules': [
                            {
                                'Name': 'prefix'|'suffix',
                                'Value': 'string'
                            },
                        ]
                    }
                }
            },
        ]
    }

    :return: Marshalled XML data
    """
    root = s3_xml.Element('NotificationConfiguration', {'xmlns': _S3_NAMESPACE})
    _add_notification_config_to_xml(
        root,
        'TopicConfiguration',
        notifications.get('TopicConfigurations', [])
    )
    _add_notification_config_to_xml(
        root,
        'QueueConfiguration',
        notifications.get('QueueConfigurations', [])
    )
    _add_notification_config_to_xml(
        root,
        'CloudFunctionConfiguration',
        notifications.get('CloudFunctionConfigurations', [])
    )

    data = io.BytesIO()
    s3_xml.ElementTree(root).write(data, encoding=None, xml_declaration=False)
    return data.getvalue()

NOTIFICATIONS_ARN_FIELDNAME_MAP = {
    'TopicConfiguration': 'Topic',
    'QueueConfiguration': 'Queue',
    'CloudFunctionConfiguration': 'CloudFunction',
}


def _add_notification_config_to_xml(node, element_name, configs):
    """
    Internal function that builds the XML sub-structure for a given
    kind of notification configuration.

    """
    for config in configs:
        config_node = s3_xml.SubElement(node, element_name)

        if 'Id' in config:
            id_node = s3_xml.SubElement(config_node, 'Id')
            id_node.text = config['Id']

        arn_node = s3_xml.SubElement(
            config_node,
            NOTIFICATIONS_ARN_FIELDNAME_MAP[element_name]
        )
        arn_node.text = config['Arn']

        for event in config['Events']:
            event_node = s3_xml.SubElement(config_node, 'Event')
            event_node.text = event

        filter_rules = config.get('Filter', {}).get(
            'Key', {}).get('FilterRules', [])
        if filter_rules:
            filter_node = s3_xml.SubElement(config_node, 'Filter')
            s3key_node = s3_xml.SubElement(filter_node, 'S3Key')
            for filter_rule in filter_rules:
                filter_rule_node = s3_xml.SubElement(s3key_node, 'FilterRule')
                name_node = s3_xml.SubElement(filter_rule_node, 'Name')
                name_node.text = filter_rule['Name']
                value_node = s3_xml.SubElement(filter_rule_node, 'Value')
                value_node.text = filter_rule['Value']
    return node


def xml_marshal_delete_objects(object_names):
    """
    Marshal Multi-Object Delete request body from object names.

    :param object_names: List of object keys to be deleted.
    :return: Serialized XML string for multi-object delete request body.
    """
    root = s3_xml.Element('Delete')

    # use quiet mode in the request - this causes the S3 Server to
    # limit its response to only object keys that had errors during
    # the delete operation.
    quiet = s3_xml.SubElement(root, 'Quiet')
    quiet.text = "true"

    # add each object to the request.
    for object_name in object_names:
        object_elt = s3_xml.SubElement(root, 'Object')
        key_elt = s3_xml.SubElement(object_elt, 'Key')
        key_elt.text = object_name

    # return the marshalled xml.
    data = io.BytesIO()
    s3_xml.ElementTree(root).write(data, encoding=None, xml_declaration=False)
    return data.getvalue()



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C) 2015 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import timedelta

import mock
from nose.tools import raises
from unittest import TestCase

from minio import Minio
from minio.error import InvalidArgumentError


class PresignedGetObjectTest(TestCase):
    @raises(TypeError)
    def test_object_is_string(self):
        client = Minio('localhost:9000')
        client.presigned_get_object('hello', 1234)

    @raises(ValueError)
    def test_object_is_not_empty_string(self):
        client = Minio('localhost:9000')
        client.presigned_get_object('hello', ' \t \n ')

    @raises(InvalidArgumentError)
    def test_expiry_limit(self):
        client = Minio('localhost:9000')
        client.presigned_get_object('hello', 'key', expires=timedelta(days=8))

    def test_can_include_response_headers(self):
        client = Minio('localhost:9000', 'my_access_key', 'my_secret_key',
                       secure=True)
        client._get_bucket_region = mock.Mock(return_value='us-east-1')
        r = client.presigned_get_object(
            'mybucket', 'myfile.pdf',
            response_headers={
                'Response-Content-Type': 'application/pdf',
                'Response-Content-Disposition': 'inline;  filename="test.pdf"'
            })
        self.assertIn('inline', r)
        self.assertIn('test.pdf', r)
        self.assertIn('application%2Fpdf', r)



# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C) 2015 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import sys

from codecs import open

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

version = ''
with open('minio/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

with open('README.md', 'r', 'utf-8') as f:
    readme = f.read()

packages = [
    'minio',
    'minio.select',
]

requires = [
    'future;python_version<"3.0"',
    'urllib3',
    'pytz',
    'certifi',
    'python-dateutil',
]

tests_requires = [
    'nose',
    'mock',
    'Faker',
]

setup(
    name='minio',
    description='MinIO Python Library for Amazon S3 Compatible Cloud Storage for Python',
    author='MinIO, Inc.',
    url='https://github.com/minio/minio-py',
    download_url='https://github.com/minio/minio-py',
    author_email='dev@min.io',
    version=version,
    long_description_content_type='text/markdown',
    package_dir={'minio': 'minio'},
    packages=packages,
    install_requires=requires,
    tests_require=tests_requires,
    license='Apache License 2.0',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    long_description=readme,
    package_data={'': ['LICENSE', 'README.md']},
    include_package_data=True,
)



# -*- coding: utf-8 -*-
# MinIO Python Library for Amazon S3 Compatible Cloud Storage, (C) 2015 MinIO, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
minio.signer
~~~~~~~~~~~~~~~

This module implements all helpers for AWS Signature version '4' support.

:copyright: (c) 2015 by MinIO, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

import collections
import hashlib
import hmac

from datetime import datetime
from .error import InvalidArgumentError
from .compat import urlsplit, queryencode
from .helpers import get_sha256_hexdigest
from .fold_case_dict import FoldCaseDict

# Signature version '4' algorithm.
_SIGN_V4_ALGORITHM = 'AWS4-HMAC-SHA256'

# Hardcoded S3 header value for X-Amz-Content-Sha256
_UNSIGNED_PAYLOAD = u'UNSIGNED-PAYLOAD'

def post_presign_signature(date, region, secret_key, policy_str):
    """
    Calculates signature version '4' for POST policy string.

    :param date: datetime formatted date.
    :param region: region of the bucket for the policy.
    :param secret_key: Amazon S3 secret access key.
    :param policy_str: policy string.
    :return: hexlified sha256 signature digest.
    """
    signing_key = generate_signing_key(date, region, secret_key)
    signature = hmac.new(signing_key, policy_str.encode('utf-8'),
                         hashlib.sha256).hexdigest()

    return signature


def presign_v4(method, url, access_key, secret_key, session_token=None,
               region=None, headers=None, expires=None, response_headers=None,
               request_date=None):
    """
    Calculates signature version '4' for regular presigned URLs.

    :param method: Method to be presigned examples 'PUT', 'GET'.
    :param url: URL to be presigned.
    :param access_key: Access key id for your AWS s3 account.
    :param secret_key: Secret access key for your AWS s3 account.
    :param session_token: Session token key set only for temporary
       access credentials.
    :param region: region of the bucket, it is optional.
    :param headers: any additional HTTP request headers to
       be presigned, it is optional.
    :param expires: final expiration of the generated URL. Maximum is 7days.
    :param response_headers: Specify additional query string parameters.
    :param request_date: the date of the request.
    """

    # Validate input arguments.
    if not access_key or not secret_key:
        raise InvalidArgumentError('Invalid access_key and secret_key.')

    if region is None:
        region = 'us-east-1'

    if headers is None:
        headers = {}

    if expires is None:
        expires = '604800'

    if request_date is None:
        request_date = datetime.utcnow()

    parsed_url = urlsplit(url)
    content_hash_hex = _UNSIGNED_PAYLOAD
    host = remove_default_port(parsed_url)
    headers['Host'] = host
    iso8601Date = request_date.strftime("%Y%m%dT%H%M%SZ")

    headers_to_sign = headers
    # Construct queries.
    query = {}
    query['X-Amz-Algorithm'] = _SIGN_V4_ALGORITHM
    query['X-Amz-Credential'] = generate_credential_string(access_key,
                                                           request_date,
                                                           region)
    query['X-Amz-Date'] = iso8601Date
    query['X-Amz-Expires'] = str(expires)
    if session_token:
        query['X-Amz-Security-Token'] = session_token

    signed_headers = get_signed_headers(headers_to_sign)
    query['X-Amz-SignedHeaders'] = ';'.join(signed_headers)

    if response_headers is not None:
        query.update(response_headers)

    # URL components.
    url_components = [parsed_url.geturl()]
    if query is not None:
        ordered_query = collections.OrderedDict(sorted(query.items()))
        query_components = []
        for component_key in ordered_query:
            single_component = [component_key]
            if ordered_query[component_key] is not None:
                single_component.append('=')
                single_component.append(
                    queryencode(ordered_query[component_key])
                )
            else:
                single_component.append('=')
            query_components.append(''.join(single_component))

        query_string = '&'.join(query_components)
        if query_string:
            url_components.append('?')
            url_components.append(query_string)
    new_url = ''.join(url_components)
    # new url constructor block ends.
    new_parsed_url = urlsplit(new_url)

    canonical_request = generate_canonical_request(method,
                                                   new_parsed_url,
                                                   headers_to_sign,
                                                   signed_headers,
                                                   content_hash_hex)
    string_to_sign = generate_string_to_sign(request_date, region,
                                             canonical_request)
    signing_key = generate_signing_key(request_date, region, secret_key)
    signature = hmac.new(signing_key, string_to_sign.encode('utf-8'),
                         hashlib.sha256).hexdigest()
    new_parsed_url = urlsplit(new_url + "&X-Amz-Signature="+signature)
    return new_parsed_url.geturl()



def get_signed_headers(headers):
    """
    Get signed headers.

    :param headers: input dictionary to be sorted.
    """
    signed_headers = []
    for header in headers:
        signed_headers.append(header.lower().strip())
    return sorted(signed_headers)


def sign_v4(method, url, region, headers=None,
            access_key=None,
            secret_key=None,
            session_token=None,
            content_sha256=None,
            request_datetime=None):
    """
    Signature version 4.

    :param method: HTTP method used for signature.
    :param url: Final url which needs to be signed.
    :param region: Region should be set to bucket region.
    :param headers: Optional headers for the method.
    :param access_key: Optional access key, if not
       specified no signature is needed.
    :param secret_key: Optional secret key, if not
       specified no signature is needed.
    :param session_token: Optional session token, set
       only for temporary credentials.
    :param content_sha256: Optional body sha256.
    :param request_datetime: Optional request date/time
    """

    # If no access key or secret key is provided return headers.
    if not access_key or not secret_key:
        return headers

    if headers is None:
        headers = FoldCaseDict()

    if region is None:
        region = 'us-east-1'

    parsed_url = urlsplit(url)
    secure = parsed_url.scheme == 'https'
    if secure:
        content_sha256 = _UNSIGNED_PAYLOAD
    if content_sha256 is None:
        # with no payload, calculate sha256 for 0 length data.
        content_sha256 = get_sha256_hexdigest('')

    host = remove_default_port(parsed_url)
    headers['Host'] = host

    if request_datetime is None:
        request_datetime = datetime.utcnow()

    headers['X-Amz-Date'] = request_datetime.strftime("%Y%m%dT%H%M%SZ")
    headers['X-Amz-Content-Sha256'] = content_sha256
    if session_token:
        headers['X-Amz-Security-Token'] = session_token

    headers_to_sign = headers

    signed_headers = get_signed_headers(headers_to_sign)
    canonical_req = generate_canonical_request(method,
                                               parsed_url,
                                               headers_to_sign,
                                               signed_headers,
                                               content_sha256)

    string_to_sign = generate_string_to_sign(request_datetime, region,
                                             canonical_req)
    signing_key = generate_signing_key(request_datetime, region, secret_key)
    signature = hmac.new(signing_key, string_to_sign.encode('utf-8'),
                         hashlib.sha256).hexdigest()

    authorization_header = generate_authorization_header(access_key,
                                                         request_datetime,
                                                         region,
                                                         signed_headers,
                                                         signature)

    headers['Authorization'] = authorization_header
    return headers


def generate_canonical_request(method, parsed_url, headers, signed_headers, content_sha256):
    """
    Generate canonical request.

    :param method: HTTP method.
    :param parsed_url: Parsed url is input from :func:`urlsplit`
    :param headers: HTTP header dictionary.
    :param content_sha256: Content sha256 hexdigest string.
    """
    # Should not encode ~. Decode it back if present.
    parsed_url_path = parsed_url.path.replace("%7E", "~")
    parsed_url_query = parsed_url.query.replace("%7E", "~")
    lines = [method, parsed_url_path, parsed_url_query]

    # Headers added to canonical request.
    header_lines = []
    for header in signed_headers:
        value = headers[header.title()]
        value = str(value).strip()
        header_lines.append(header + ':' + str(value))

    lines = lines + header_lines
    lines.append('')

    lines.append(';'.join(signed_headers))
    lines.append(content_sha256)
    return '\n'.join(lines)


def generate_string_to_sign(date, region, canonical_request):
    """
    Generate string to sign.

    :param date: Date is input from :meth:`datetime.datetime`
    :param region: Region should be set to bucket region.
    :param canonical_request: Canonical request generated previously.
    """
    formatted_date_time = date.strftime("%Y%m%dT%H%M%SZ")

    canonical_request_hasher = hashlib.sha256()
    canonical_request_hasher.update(canonical_request.encode('utf-8'))
    canonical_request_sha256 = canonical_request_hasher.hexdigest()
    scope = generate_scope_string(date, region)

    return '\n'.join([_SIGN_V4_ALGORITHM,
                      formatted_date_time,
                      scope,
                      canonical_request_sha256])


def generate_signing_key(date, region, secret_key):
    """
    Generate signing key.

    :param date: Date is input from :meth:`datetime.datetime`
    :param region: Region should be set to bucket region.
    :param secret_key: Secret access key.
    """
    formatted_date = date.strftime("%Y%m%d")

    key1_string = 'AWS4' + secret_key
    key1 = key1_string.encode('utf-8')
    key2 = hmac.new(key1, formatted_date.encode('utf-8'),
                    hashlib.sha256).digest()
    key3 = hmac.new(key2, region.encode('utf-8'), hashlib.sha256).digest()
    key4 = hmac.new(key3, 's3'.encode('utf-8'), hashlib.sha256).digest()

    return hmac.new(key4, 'aws4_request'.encode('utf-8'),
                    hashlib.sha256).digest()


def generate_scope_string(date, region):
    """
    Generate scope string.

    :param date: Date is input from :meth:`datetime.datetime`
    :param region: Region should be set to bucket region.
    """
    formatted_date = date.strftime("%Y%m%d")
    scope = '/'.join([formatted_date,
                      region,
                      's3',
                      'aws4_request'])
    return scope


def generate_credential_string(access_key, date, region):
    """
    Generate credential string.

    :param access_key: Server access key.
    :param date: Date is input from :meth:`datetime.datetime`
    :param region: Region should be set to bucket region.
    """
    return access_key + '/' + generate_scope_string(date, region)


def generate_authorization_header(access_key, date, region,
                                  signed_headers, signature):
    """
    Generate authorization header.

    :param access_key: Server access key.
    :param date: Date is input from :meth:`datetime.datetime`
    :param region: Region should be set to bucket region.
    :param signed_headers: Signed headers.
    :param signature: Calculated signature.
    """
    signed_headers_string = ';'.join(signed_headers)
    credential = generate_credential_string(access_key, date, region)
    auth_header = [_SIGN_V4_ALGORITHM, 'Credential=' + credential + ',',
                   'SignedHeaders=' + signed_headers_string + ',',
                   'Signature=' + signature]
    return ' '.join(auth_header)

def remove_default_port(parsed_url):
    default_ports = {
        'http': 80,
        'https': 443
    }
    if any(parsed_url.scheme == scheme and parsed_url.port == port
           for scheme, port in default_ports.items()):
        # omit default port (i.e. 80 or 443)
        host = parsed_url.hostname
    else:
        host = parsed_url.netloc
    return host
